# coding=utf-8
# Filename: test_core.py
# pylint: disable=C0111,E1003,R0904,C0103,R0201,C0102
from __future__ import division, absolute_import, print_function

__author__ = 'tamasgal'

from km3pipe.testing import TestCase, StringIO, MagicMock
from km3pipe.core import Pipeline, Module, Pump, Blob


class TestPipeline(TestCase):
    """Tests for the main pipeline"""

    def setUp(self):
        self.pl = Pipeline()

    def test_attach(self):
        self.pl.attach(Module, 'module1')
        self.pl.attach(Module, 'module2')
        self.assertEqual('module1', self.pl.modules[0].name)
        self.assertEqual('module2', self.pl.modules[1].name)

    def test_drain_calls_process_method_on_each_attached_module(self):
        pl = Pipeline(blob=1)
        pl.attach(Module, 'module1')
        pl.attach(Module, 'module2')
        for module in pl.modules:
            module.process = MagicMock(return_value=1)
        pl.drain(1)
        for module in pl.modules:
            module.process.assert_called_once_with(1)

    def test_finish(self):
        self.pl.finish()

    def test_drain_calls_finish_on_each_attached_module(self):
        self.pl.attach(Module, 'module1')
        self.pl.attach(Module, 'module2')
        for module in self.pl.modules:
            module.finish = MagicMock()
        self.pl.drain(4)
        for module in self.pl.modules:
            module.finish.assert_called_once_with()


class TestModule(TestCase):
    """Tests for the pipeline module"""

    def test_name_can_be_set_on_init(self):
        name = 'foo'
        module = Module(name=name)
        self.assertEqual(name, module.name)

    def test_name_is_read_only(self):
        module = Module(name='foo')
        with self.assertRaises(AttributeError):
            module.name = 'narf'

    def test_process(self):
        blob = Blob()
        module = Module(name='foo')
        processed_blob = module.process(blob)
        self.assertIs(blob, processed_blob)

    def test_add_parameter(self):
        module = Module()
        module.add('foo', 'default')
        self.assertDictEqual({'foo': 'default'}, module.parameters)

    def test_get_parameter(self):
        module = Module()
        module.add('foo', 'default')
        self.assertEqual('default', module.get('foo'))

    def test_default_parameter_value_can_be_overwritten(self):
        class Foo(Module):
            def __init__(self, **context):
                super(self.__class__, self).__init__(**context)
                self.foo = self.get('foo') or 'default_foo'
        module = Foo()
        self.assertEqual('default_foo', module.foo)
        module = Foo(foo='overwritten')
        self.assertEqual('overwritten', module.foo)

    def test_finish(self):
        module = Module()
        module.finish()


class TestPump(TestCase):
    """Tests for the pump"""

    def test_rewind_file(self):
        pump = Pump()
        test_file = StringIO("Some content")
        pump.blob_file = test_file
        pump.blob_file.read(1)
        self.assertEqual(1, pump.blob_file.tell())
        pump.rewind_file()
        self.assertEqual(0, pump.blob_file.tell())


class TestBlob(TestCase):
    """Tests for the blob holding the data"""

    def test_field_can_be_added(self):
        blob = Blob()
        blob['foo'] = 1
        self.assertEqual(1, blob['foo'])



#!/usr/bin/env python
# coding=utf-8
# Filename: jpp.py
# pylint: disable=
"""
Pump for the jpp file read through aanet interface.

"""
from __future__ import division, absolute_import, print_function

import socket

from km3pipe import Pump
from km3pipe.logger import logging
import threading
try:
    from Queue import Queue, Empty
except ImportError:
    from queue import Queue, Empty

log = logging.getLogger(__name__)  # pylint: disable=C0103


class CHPump(Pump):
    """A pump for ControlHost data."""
    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)

        self.host = self.get('host') or '127.0.0.1'
        self.port = self.get('port') or 5553
        self.tags = self.get('tags') or "MSG"
        self.timeout = self.get('timeout') or 60*60*24
        self.max_queue = self.get('max_queue') or 50
        self.key_for_data = self.get('key_for_data') or 'CHData'
        self.key_for_prefix = self.get('key_for_prefix') or 'CHPrefix'

        self.queue = Queue()
        self.client = None
        self.thread = None

        print("Connecting to {0} on port {1}\n"
              "Subscribed tags: {2}\n"
              "Connection timout: {3}s\n"
              "Maximum queue size for incoming data: {4}"
              .format(self.host, self.port, self.tags, self.timeout,
                      self.max_queue))

        self._init_controlhost()
        self._start_thread()

    def _start_thread(self):
        self.thread = threading.Thread(target=self._run, args=())
        self.thread.daemon = True
        self.thread.start()

    def _init_controlhost(self):
        """Set up the controlhost connection"""
        from controlhost import Client
        self.client = Client(self.host, self.port)
        self.client._connect()
        for tag in self.tags.split(','):
            self.client.subscribe(tag.strip())

    def _run(self):
        while True:
            prefix, data = self.client.get_message()
            if self.queue.qsize() > self.max_queue:
                log.warn("Maximum queue size ({0}) reached, dropping data."
                         .format(self.max_queue))
            else:
                self.queue.put((prefix, data))

    def process(self, blob):
        """Wait for the next packet and put it in the blob"""
        try:
            prefix, data = self.queue.get(timeout=self.timeout)
        except Empty:
            log.warn("ControlHost timeout ({0}s) exceeded".format(self.timeout))
            raise StopIteration("ControlHost timeout exceeded.")
        blob[self.key_for_prefix] = prefix
        blob[self.key_for_data] = data
        return blob

    def finish(self):
        """Clean up the JLigier controlhost connection"""
        self.client._disconnect()



# coding=utf-8
# Filename: core.py
# pylint: disable=locally-disabled
"""
The core of the KM3Pipe framework.

"""
from __future__ import division, absolute_import, print_function

__author__ = 'tamasgal'

import signal

from km3pipe.hardware import Detector

import logging
from km3pipe.logger import logging

log = logging.getLogger(__name__)  # pylint: disable=C0103


class Pipeline(object):
    """The holy pipeline which holds everything together"""

    def __init__(self, blob=None):
        self.modules = []
        self.geometry = None
        self.blob = blob or Blob()
        self._cycle_count = 0
        self._stop = False
        self._finished = False

    def attach(self, module_class, name=None, **kwargs):
        """Attach a module to the pipeline system"""
        if not name:
            name = module_class.__name__
        module = module_class(name=name, **kwargs)
        log.info("Attaching module '{0}'".format(name))
        try:
            module.get_detector()
            self.geometry = module
        except AttributeError:
            if len(self.modules) < 1 and not isinstance(module, Pump):
                log.error("The first module to attach to the pipeline should "
                          "be a Pump!")
            self.modules.append(module)

    def _drain(self, cycles=None):
        """Activate the pump and let the flow go.

        This will call the process() method on each attached module until
        a StopIteration is raised, usually by a pump when it reached the EOF.

        A StopIteration is also raised when self.cycles was set and the
        number of cycles has reached that limit.

        """
        if not cycles:
            log.info("No cycle count set, the pipeline may be drained forever.")

        if self.geometry:
            log.info("Setting up the detector geometry.")
            for module in self.modules:
                module.detector = self.geometry.get_detector()

        try:
            while not self._stop:
                self._cycle_count += 1
                log.debug("Pumping blob #{0}".format(self._cycle_count))
                self.blob = self.modules[0].process(self.blob)
                for module in self.modules[1:]:
                    if self.blob is None:
                        log.debug("Skipping {0}, due to empty blob."
                                  .format(module.name))
                        continue
                    log.debug("Processing {0} ".format(module.name))
                    self.blob = module.process(self.blob)
                if cycles and self._cycle_count >= cycles:
                    raise StopIteration
        except StopIteration:
            log.info("Nothing left to pump through.")
        self.finish()

    def drain(self, cycles=None):
        """Execute _drain while trapping KeyboardInterrupt"""
        log.info("Now draining...")
        signal.signal(signal.SIGINT, self._handle_ctrl_c)
        try:
            self._drain(cycles)
        except KeyboardInterrupt:
            pass

    def finish(self):
        """Call finish() on each attached module"""
        for module in self.modules:
            log.info("Finishing {0}".format(module.name))
            module.pre_finish()
        self._finished = True

    def _handle_ctrl_c(self, *args):
        """Handle the keyboard interrupts."""
        if self._stop:
            print("Forced shutdown...")
            raise SystemExit
        if not self._stop:
            print(42*'=' + "\nGot CTRL+C, waiting for the current cycle...\n"
                  "Press CTRL+C again if you're in hurry!\n" + 42*'=')
            self._stop = True


class Module(object):
    """The module which can be attached to the pipeline"""

    def __init__(self, name=None, **parameters):
        log.debug("Initialising {0}".format(name))
        self._name = name
        self.parameters = parameters
        self.detector = None

    @property
    def name(self):
        """The name of the module"""
        return self._name

    def add(self, name, value):
        """Add the parameter with the desired value to the dict"""
        self.parameters[name] = value

    def get(self, name):
        """Return the value of the requested parameter"""
        return self.parameters.get(name)

    def process(self, blob):  # pylint: disable=R0201
        """Knead the blob and return it"""
        return blob

    def finish(self):
        """Clean everything up."""
        pass

    def pre_finish(self):
        """Do the last few things before calling finish()"""
        self.finish()


class Pump(Module):
    """The pump with basic file or socket handling."""

    def __init__(self, **context):
        Module.__init__(self, **context)
        self.blob_file = None

    def open_file(self, filename):
        """Open the file with filename"""
        try:
            self.blob_file = open(filename, 'rb')
        except TypeError:
            log.error("Please specify a valid filename.")
            raise SystemExit
        except IOError as error_message:
            log.error(error_message)
            raise SystemExit

    def process(self, blob):
        """Create a blob"""
        raise NotImplementedError("The pump has no process() method!")

    def rewind_file(self):
        """Put the file pointer to position 0"""
        self.blob_file.seek(0, 0)

    def close_file(self):
        """Close file."""
        if self.blob_file:
            self.blob_file.close()

    def pre_finish(self):
        """Clean up open file or socket-handlers."""
        Module.finish(self)
        self.close_file()


class Blob(dict):
    """A simple dict with a fancy name. This should hold the data."""
    pass


class Geometry(Module):
    """A very simple, preliminary Module which gives access to the geometry"""
    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)
        filename = self.get('filename')
        self.detector = Detector(filename)

    def get_detector(self):
        """Return the detector"""
        return self.detector




# coding=utf-8
# Filename: decorators.py
# pylint: disable=locally-disabled
"""
Function decorators.

"""
from __future__ import division, absolute_import, print_function

__author__ = 'tamasgal'


def remain_file_pointer(function):
    """Remain the file pointer position after calling the decorated function

    This decorator assumes that the last argument is the file handler.

    """
    def wrapper(*args, **kwargs):
        """Wrap the function and remain its parameters and return values"""
        file_obj = args[-1]
        old_position = file_obj.tell()
        return_value = function(*args, **kwargs)
        file_obj.seek(old_position, 0)
        return return_value
    return wrapper




# coding=utf-8
# Filename: __init__.py
"""
Common unit testing support for km3pipe.

"""
from __future__ import division, absolute_import, print_function

try:
    from unittest2 import TestCase, skip, skipIf
except ImportError:
    from unittest import TestCase, skip, skipIf

try:
    from cStringIO import StringIO
except ImportError:
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO

from mock import MagicMock



# coding=utf-8
# Filename: dataclasses.py
# pylint: disable=W0232,C0103,C0111
"""
...

"""
from __future__ import division, absolute_import, print_function

__all__ = ('Point', 'Position', 'Direction')


import numpy as np

from km3pipe.tools import angle_between

class Point(np.ndarray):
    """Represents a point in a 3D space"""
    def __new__(cls, input_array=(np.nan, np.nan, np.nan)):
        """Add x, y and z to the ndarray"""
        obj = np.asarray(input_array).view(cls)
        return obj

    @property
    def x(self):
        return self[0]

    @x.setter
    def x(self, value):
        self[0] = value

    @property
    def y(self):
        return self[1]

    @y.setter
    def y(self, value):
        self[1] = value

    @property
    def z(self):
        return self[2]

    @z.setter
    def z(self, value):
        self[2] = value


class Position(Point):
    """Represents a point in a 3D space"""
    pass


class Direction(Point):
    """Represents a direction in a 3D space

    The direction vector always normalises itself when an attribute is changed.

    """
    def __new__(cls, input_array=(1, 0, 0)):
        """Add x, y and z to the ndarray"""
        normed_array = np.array(input_array) / np.linalg.norm(input_array)
        obj = np.asarray(normed_array).view(cls)
        return obj

    def _normalise(self):
        normed_array = self / np.linalg.norm(self)
        self[0] = normed_array[0]
        self[1] = normed_array[1]
        self[2] = normed_array[2]

    @property
    def x(self):
        return self[0]

    @x.setter
    def x(self, value):
        self[0] = value
        self._normalise()

    @property
    def y(self):
        return self[1]

    @y.setter
    def y(self, value):
        self[1] = value
        self._normalise()

    @property
    def z(self):
        return self[2]

    @z.setter
    def z(self, value):
        self[2] = value
        self._normalise()

    @property
    def zenith(self):
        return angle_between(self, (0, 0, -1))

    def __str__(self):
        return "({0:.4}, {1:.4}, {2:.4})".format(self.x, self.y, self.z)





# coding=utf-8
# Filename: aanet.py
# pylint: disable=locally-disabled
"""
Pump for the Aanet data format.

"""
from __future__ import division, absolute_import, print_function

from km3pipe import Pump
from km3pipe.logger import logging

log = logging.getLogger(__name__)  # pylint: disable=C0103


class AanetPump(Pump):
    """A pump for binary Aanet files."""

    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)

        # pylint: disable:F0401,W0612
        import aa
        from ROOT import TFile, Evt, Trk

        self.filename = self.get('filename')
        self.treename = self.get('treename') or "E"
        if not self.filename:
            raise ValueError("No filename defined")
        self.index = 0
        self.rootfile = TFile(self.filename)
        self.evt = Evt()
        self.E = self.rootfile.Get(self.treename)
        self.N = self.E.GetEntries()
        self.E.SetBranchAddress('Evt', self.evt)

    def get_blob(self, index):
        self.E.GetEntry(index)
        return {'Evt': self.evt,
                'hits': self.evt.hits}

    def process(self, blob):
        self.E.GetEntry(self.index)
        if self.index == self.N:
            raise StopIteration
        else:
            self.index += 1
            return {'Evt': self.evt}



# coding=utf-8
# Filename: widgets.py
"""
GUI elements for the pipeinspector.

"""
from __future__ import division, absolute_import, print_function

import urwid
import pprint
import math
import sys

from pipeinspector.settings import UI

class BlobBrowser(urwid.Frame):
    def __init__(self):
        self.items = []
        self.cursor_position = 0

        self.header = urwid.AttrMap(urwid.Text('Keys:'), 'head')

        self.listbox = urwid.ListBox(urwid.SimpleListWalker(self.items))
        self.frame = urwid.Frame(self.listbox, header=self.header)
        line_box = urwid.AttrMap(urwid.LineBox(self.frame), 'body')
        urwid.Frame.__init__(self, line_box)
        self.overlay = None
        self.popup = None

    def load(self, blob):
        del self.listbox.body[:]
        new_items = []
        for key in sorted(blob.keys()):
            item_widget = ItemWidget(key, blob[key])
            new_items.append(item_widget)
            urwid.connect_signal(item_widget, 'key_selected', self.key_selected)
        self.listbox.body.extend(new_items)
        self.listbox.set_focus(self.cursor_position)

    def key_selected(self, data):
        def formatter(obj):
            if hasattr(obj, '__insp__'):
                return obj.__insp__()
            if hasattr(obj, 'size'):
                output = ""
                for obj in data:
                    output += str(obj) + "\n"
                return output
            return pprint.pformat(obj)

        content = [urwid.Text(line) for line in formatter(data).split('\n')]
        self.popup = urwid.ListBox(content)
        popup_box = urwid.LineBox(self.popup)
        self.overlay = urwid.Overlay(popup_box, self.body,
                                  'center',
                                  ('relative', 80),
                                  'middle',
                                  ('relative', 80),)
        self.body = self.overlay

    def keypress(self, size, key):
        input = urwid.Frame.keypress(self, size, key)
        self.cursor_position = self.listbox.focus_position
        if self.overlay:
            if input in UI.keys['escape']:
                self.body = self.overlay.bottom_w
                self.overlay = None
        else:
            return input


class ItemWidget (urwid.WidgetWrap):
    signals = ['key_selected']
    def __init__ (self, key, data):
        self.key = key
        self.data = data
        self.item = [
            ('fixed', 35, urwid.Padding(
                urwid.AttrWrap(urwid.Text(key), 'body', 'focus'), left=2)),
            urwid.AttrWrap(urwid.Text(str(type(data))), 'body', 'focus'),
            #urwid.AttrWrap(urwid.Text(str(len(data))), 'body', 'focus'),
            urwid.AttrWrap(urwid.Text(str(sys.getsizeof(data))), 'body', 'focus'),
        ]
        w = urwid.Columns(self.item)
        self.__super.__init__(w)

    def selectable(self):
        return True

    def keypress(self, size, key):
        if key == 'x':
            urwid.emit_signal(self, 'key_selected', self.data)
        return key


class BlobWidget(urwid.Pile):
    signals = ['blob_selected']
    def __init__(self):
        self.width = 50
        self.size = (0,)
        self.index = 0
        urwid.Pile.__init__(self, [urwid.Text('', wrap='clip'),
                                   urwid.Text('', wrap='clip'),
                                   urwid.Text('', wrap='clip')])

    def goto_blob(self, position):
        self.index = position
        self._emit_blob_selected()
        self.draw()

    def previous_blob(self, step=1):
        self.index -= step
        if self.index <= 0:
            self.index = 0
        self._emit_blob_selected()
        self.draw()

    def next_blob(self, step=1):
        self.index += step
        self._emit_blob_selected()
        self.draw()

    def draw(self):
        self.widget_list[0].set_text(self._make_blob_icons(self.index))
        self.widget_list[1].set_text(self._make_ruler(self.index))
        self.widget_list[2].set_text(self._make_scale_labels(self.index))

    def render(self, size, focus):
        self.size = size
        self.draw()
        return urwid.Pile.render(self, size, focus)

    def _emit_blob_selected(self):
        urwid.emit_signal(self, 'blob_selected', self.index)

    def _make_ruler(self, start):
        if start <= 10:
            start = 0
        else:
            start -= 10
        segment = "|    '    "
        repeat = int(math.ceil(self.width / len(segment)) + 1)
        ruler = segment * repeat
        slice_start = (start % len(segment))
        slice_end = (start % len(segment)) + self.width
        return ruler[slice_start:slice_end]

    def _make_scale_labels(self, start):
        if start <= 10:
            start = 0
        else:
            start -= 10
        lowest_tick = int(math.floor(start / 10) * 10)
        highest_tick = lowest_tick + self.width
        ticks_labels = ['{0:<10}'.format(i)
                 for i in range(lowest_tick, highest_tick, 10)]
        slice_start = (start % 10)
        slice_end = (start % 10) + self.width
        ticks = ''.join(ticks_labels)[slice_start:slice_end]
        return ticks

    def _make_blob_icons(self, start):
        icon = 'B'
        if start < 10:
            icons = '.' + icon * (self.width - 1)
        else:
            icons = icon * self.width
        if start > 10:
            start = 10
        return [('blob', icons[:start]),
                ('blob_selected', icons[start]),
                ('blob', icons[start + 1:])]





# coding=utf-8
# Filename: test_hardware.py
# pylint: disable=C0111,C0103,R0904
"""
Detector description (detx format v5)

global_det_id ndoms
dom_id line_id floor_id npmts
 pmt_id_global x y z dx dy dz t0
 pmt_id_global x y z dx dy dz t0
 ...
 pmt_id_global x y z dx dy dz t0
dom_id line_id floor_id npmts
 ...

"""
from __future__ import division, absolute_import, print_function

from km3pipe.testing import TestCase, StringIO, skipIf
from km3pipe.hardware import Detector, PMT


EXAMPLE_DETX = StringIO("\n".join((
    "1 3",
    "1 1 1 3",
    " 1 1.1 1.2 1.3 -1.1  0.2  0.3 10",
    " 2 1.4 1.5 1.6  0.1 -1.2  0.3 20",
    " 3 1.7 1.8 1.9  0.1  0.2 -1.3 30",
    "2 1 2 3",
    " 4 2.1 2.2 2.3 -1.1  0.2  0.3 40",
    " 5 2.4 2.5 2.6  0.1 -1.2  0.3 50",
    " 6 2.7 2.8 2.9  0.1  0.2 -1.3 60",
    "3 1 3 3",
    " 7 3.1 3.2 3.3 -1.1  0.2  0.3 70",
    " 8 3.4 3.5 3.6  0.1 -1.2  0.3 80",
    " 9 3.7 3.8 3.9  0.1  0.2 -1.3 90",)))

EXAMPLE_DETX_MIXED_IDS = StringIO("\n".join((
    "1 3",
    "8 1 1 3",
    " 83 1.1 1.2 1.3 -1.1  0.2  0.3 10",
    " 81 1.4 1.5 1.6  0.1 -1.2  0.3 20",
    " 82 1.7 1.8 1.9  0.1  0.2 -1.3 30",
    "7 1 2 3",
    " 71 2.1 2.2 2.3 -1.1  0.2  0.3 40",
    " 73 2.4 2.5 2.6  0.1 -1.2  0.3 50",
    " 72 2.7 2.8 2.9  0.1  0.2 -1.3 60",
    "6 1 3 3",
    " 62 3.1 3.2 3.3 -1.1  0.2  0.3 70",
    " 63 3.4 3.5 3.6  0.1 -1.2  0.3 80",
    " 61 3.7 3.8 3.9  0.1  0.2 -1.3 90",)))

EXAMPLE_DETX_WRITE = StringIO("\n".join((
    "1 3",
    "1 1 1 3",
    " 1 1.1 1.2 1.3 1.0 0.0 0.0 10.0",
    " 2 1.4 1.5 1.6 0.0 1.0 0.0 20.0",
    " 3 1.7 1.8 1.9 0.0 0.0 1.0 30.0",
    "2 1 2 3",
    " 4 2.1 2.2 2.3 0.0 1.0 0.0 40.0",
    " 5 2.4 2.5 2.6 0.0 0.0 1.0 50.0",
    " 6 2.7 2.8 2.9 1.0 0.0 0.0 60.0",
    "3 1 3 3",
    " 7 3.1 3.2 3.3 0.0 0.0 1.0 70.0",
    " 8 3.4 3.5 3.6 0.0 1.0 0.0 80.0",
    " 9 3.7 3.8 3.9 1.0 0.0 0.0 90.0\n",)))

EXAMPLE_MC_DETX_WRITE_MIXED_IDS = StringIO("\n".join((
    "-1 3",
    "6 1 1 3",
    " 31 1.1 1.2 1.3 1.0 0.0 0.0 10.0",
    " 22 1.4 1.5 1.6 0.0 1.0 0.0 20.0",
    " 13 1.7 1.8 1.9 0.0 0.0 1.0 30.0",
    "3 1 2 3",
    " 34 2.1 2.2 2.3 1.0 0.0 0.0 40.0",
    " 45 2.4 2.5 2.6 0.0 1.0 0.0 50.0",
    " 16 2.7 2.8 2.9 0.0 0.0 1.0 60.0",
    "9 1 3 3",
    " 17 3.1 3.2 3.3 1.0 0.0 0.0 70.0",
    " 48 3.4 3.5 3.6 0.0 1.0 0.0 80.0",
    " 39 3.7 3.8 3.9 0.0 0.0 1.0 90.0\n",)))

class TestDetector(TestCase):

    def setUp(self):
        self.det = Detector()
        self.det.det_file = EXAMPLE_DETX

    def test_parse_header_extracts_correct_det_id(self):
        self.det.parse_header()
        self.assertEqual(1, self.det.det_id)

    def test_parse_header_extracts_correct_n_doms(self):
        self.det.parse_header()
        self.assertEqual(3, self.det.n_doms)

    def test_parse_doms_maps_each_dom_correctly(self):
        self.det.parse_doms()
        expected = {1: (1, 1, 3), 2: (1, 2, 3), 3: (1, 3, 3)}
        self.assertDictEqual(expected, self.det.doms)

    def test_parse_doms_maps_each_dom_correctly_for_mixed_pmt_ids(self):
        self.det.det_file = EXAMPLE_DETX_MIXED_IDS
        self.det.parse_doms()
        expected = {8: (1, 1, 3), 7: (1, 2, 3), 6: (1, 3, 3)}
        self.assertDictEqual(expected, self.det.doms)

    @skipIf(True, "Weird one hour bias on date?")
    def test_parse_doms_fills_pmts_dict(self):
        self.det.parse_doms()
        self.assertEqual(9, len(self.det.pmts))
        self.assertTupleEqual((7, 3.1, 3.2, 3.3, -1.1, 0.2, 0.3, 70),
                              self.det.pmts[(1, 3, 0)])

    def test_dom_positions(self):
        self.det.parse_doms()
        for i, position in enumerate(self.det.dom_positions):
            self.assertAlmostEqual(i + 1.1, position.x)
            self.assertAlmostEqual(i + 1.2, position.y)
            self.assertAlmostEqual(i + 1.3, position.z)

    def test_omkeys(self):
        self.det.parse_doms()
        self.assertEqual((1, 1, 0), self.det.pmt_with_id(1).omkey)
        self.assertEqual((1, 2, 1), self.det.pmt_with_id(5).omkey)

    def test_pmt_with_id_raises_exception_for_invalid_id(self):
        self.det.parse_doms()
        with self.assertRaises(KeyError):
            self.det.pmt_with_id(100)

    @skipIf(True, "DOM positions ordering unclear")
    def test_dom_positions_with_mixed_pmt_ids(self):
        self.det.det_file = EXAMPLE_DETX_MIXED_IDS
        self.det.parse_doms()
        for i, position in enumerate(self.det.dom_positions):
            self.assertAlmostEqual(i + 1.1, position.x)
            self.assertAlmostEqual(i + 1.2, position.y)
            self.assertAlmostEqual(i + 1.3, position.z)

    @skipIf(True, "DOM ordering is probably not important!")
    def test_ascii_detector(self):
        self.det.det_file = EXAMPLE_MC_DETX_WRITE_MIXED_IDS
        self.det.parse_header()
        self.det.parse_doms()
        self.assertEqual(self.det.det_file.getvalue(), self.det.ascii)

    def test_pmtid2omkey_old(self):
        pmtid2omkey = self.det.pmtid2omkey_old
        self.assertEqual((1, 13, 12), tuple(pmtid2omkey(168)))
        self.assertEqual((1, 12, 18), tuple(pmtid2omkey(205)))
        self.assertEqual((1, 11, 22), tuple(pmtid2omkey(240)))
        self.assertEqual((4, 11, 2), tuple(pmtid2omkey(1894)))
        self.assertEqual((9, 18, 0), tuple(pmtid2omkey(4465)))
        self.assertEqual((95, 7, 16), tuple(pmtid2omkey(52810)))
        self.assertEqual((95, 4, 13), tuple(pmtid2omkey(52900)))

    def test_pmtid2omkey_old_handles_floats(self):
        pmtid2omkey = self.det.pmtid2omkey_old
        self.assertEqual((1, 13, 12), tuple(pmtid2omkey(168.0)))
        self.assertEqual((1, 12, 18), tuple(pmtid2omkey(205.0)))
        self.assertEqual((1, 11, 22), tuple(pmtid2omkey(240.0)))
        self.assertEqual((4, 11, 2), tuple(pmtid2omkey(1894.0)))
        self.assertEqual((9, 18, 0), tuple(pmtid2omkey(4465.0)))
        self.assertEqual((95, 7, 16), tuple(pmtid2omkey(52810.0)))
        self.assertEqual((95, 4, 13), tuple(pmtid2omkey(52900.0)))






class TestPMT(TestCase):

    def test_init(self):
        pmt = PMT(1, (1, 2, 3), (4, 5, 6), 7, 8, (9, 10, 11))
        self.assertEqual(1, pmt.id)
        self.assertEqual(1, pmt.pos.x)
        self.assertEqual(2, pmt.pos.y)
        self.assertEqual(3, pmt.pos.z)
        self.assertAlmostEqual(0.455842, pmt.dir.x, 6)
        self.assertAlmostEqual(0.569803, pmt.dir.y, 6)
        self.assertAlmostEqual(0.683763, pmt.dir.z, 6)
        self.assertEqual(7, pmt.t0)
        self.assertEqual(8, pmt.daq_channel)
        self.assertEqual((9, 10, 11), pmt.omkey)



# coding=utf-8
# Filename: daq.py
# pylint: disable=R0903
"""
Pumps for the DAQ data formats.

"""
from __future__ import division, absolute_import, print_function

import struct
from struct import unpack
import pprint

from km3pipe import Pump, Blob
from km3pipe.logger import logging

log = logging.getLogger(__name__)  # pylint: disable=C0103

DATA_TYPES = {
    101: 'DAQSuperFrame',
    201: 'DAQSummaryFrame',
    1001: 'DAQTimeslice',
    2001: 'DAQSummaryslice',
    10001: 'DAQEvent',
}


class DAQPump(Pump):
    """A pump for binary DAQ files."""

    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)
        self.filename = self.get('filename')
        self.frame_positions = []

        if self.filename:
            self.open_file(self.filename)
            self.determine_frame_positions()
        else:
            log.warn("No filename specified. Take care of the file handling!")

    def next_blob(self):
        """Get the next frame from file"""
        blob_file = self.blob_file
        try:
            preamble = DAQPreamble(file_obj=blob_file)
        except struct.error:
            raise StopIteration

        try:
            data_type = DATA_TYPES[preamble.data_type]
        except KeyError:
            data_type = 'Unknown'

        blob = Blob()
        blob[data_type] = None
        blob['DAQPreamble'] = preamble

        if data_type == 'DAQSummaryslice':
            daq_frame = DAQSummaryslice(blob_file)
            blob[data_type] = daq_frame
            blob['DAQHeader'] = daq_frame.header
        elif data_type == 'DAQEvent':
            daq_frame = DAQEvent(blob_file)
            blob[data_type] = daq_frame
            blob['DAQHeader'] = daq_frame.header
        else:
            log.warning("Skipping DAQ frame with data type code '{0}'."
                        .format(preamble.data_type))
            blob_file.seek(preamble.length - DAQPreamble.size, 1)

        return blob

    def seek_to_frame(self, index):
        """Move file pointer to the frame with given index."""
        pointer_position = self.frame_positions[index]
        self.blob_file.seek(pointer_position, 0)

    def get_blob(self, index):
        """Return blob at given index."""
        self.seek_to_frame(index)
        return self.next_blob()

    def determine_frame_positions(self):
        """Record the file pointer position of each frame"""
        self.rewind_file()
        try:
            while True:
                pointer_position = self.blob_file.tell()
                length = struct.unpack('<i', self.blob_file.read(4))[0]
                self.blob_file.seek(length - 4, 1)
                self.frame_positions.append(pointer_position)
        except struct.error:
            pass
        self.rewind_file()
        log.info("Found {0} frames.".format(len(self.frame_positions)))

    def process(self, blob):
        """Pump the next blob to the modules"""
        return self.next_blob()

    def finish(self):
        """Clean everything up"""
        self.blob_file.close()


class DAQPreamble(object):
    """Wrapper for the JDAQPreamble binary format.

    Args:
      file_obj (file): The binary file, where the file pointer is at the
        beginning of the header.

    Attributes:
      size (int): The size of the original DAQ byte representation.
      data_type (int): The data type of the following frame. The coding is
        stored in the ``DATA_TYPES`` dictionary::

            101: 'DAQSuperFrame'
            201: 'DAQSummaryFrame'
            1001: 'DAQTimeslice'
            2001: 'DAQSummaryslice'
            10001: 'DAQEvent'

    """
    size = 8

    def __init__(self, byte_data=None, file_obj=None):
        self.length = None
        self.data_type = None
        if byte_data:
            self._parse_byte_data(byte_data)
        if file_obj:
            self._parse_file(file_obj)

    def _parse_byte_data(self, byte_data):
        """Extract the values from byte string."""
        self.length, self.data_type = unpack('<ii', byte_data[:self.size])

    def _parse_file(self, file_obj):
        """Directly read from file handler.

        Note that this will move the file pointer.

        """
        byte_data = file_obj.read(self.size)
        self._parse_byte_data(byte_data)

    def __repr__(self):
        description = "Length: {0}\nDataType: {1}"\
            .format(self.length, self.data_type)
        return description


class DAQHeader(object):
    """Wrapper for the JDAQHeader binary format.

    Args:
      file_obj (file): The binary file, where the file pointer is at the
        beginning of the header.

    Attributes:
      size (int): The size of the original DAQ byte representation.

    """
    size = 16

    def __init__(self, byte_data=None, file_obj=None):
        self.run = None
        self.time_slice = None
        self.time_stamp = None
        if byte_data:
            self._parse_byte_data(byte_data)
        if file_obj:
            self._parse_file(file_obj)

    def _parse_byte_data(self, byte_data):
        """Extract the values from byte string."""
        run, time_slice, time_stamp = unpack('<iiQ', byte_data[:self.size])
        self.run = run
        self.time_slice = time_slice
        self.time_stamp = time_stamp

    def _parse_file(self, file_obj):
        """Directly read from file handler.

        Note:
          This will move the file pointer.

        """
        byte_data = file_obj.read(self.size)
        self._parse_byte_data(byte_data)

    def __repr__(self):
        description = "Run: {0}\nTime slice: {1}\nTime stamp: {2}"\
                      .format(self.run, self.time_slice, self.time_stamp)
        return description


class DAQSummaryslice(object):
    """Wrapper for the JDAQSummarySlice binary format.

    Args:
      file_obj (file): The binary file, where the file pointer is at the
        beginning of the header.

    Attributes:
      n_summary_frames (int): The number of summary frames.
      summary_frames (dict): The PMT rates for each DOM. The key is the DOM
        identifier and the corresponding value is a sorted list of PMT rates.

    """
    def __init__(self, file_obj):
        self.header = DAQHeader(file_obj=file_obj)
        self.n_summary_frames = unpack('<i', file_obj.read(4))[0]
        self.summary_frames = {}

        self._parse_summary_frames(file_obj)

    def _parse_summary_frames(self, file_obj):
        """Iterate through the byte data and fill the summary_frames"""
        for _ in range(self.n_summary_frames):
            dom_id = unpack('<i', file_obj.read(4))[0]
            pmt_rates = unpack('b'*31, file_obj.read(31))
            self.summary_frames[dom_id] = pmt_rates


class DAQEvent(object):
    """Wrapper for the JDAQEvent binary format.

    Args:
      file_obj (file): The binary file, where the file pointer is at the

    Attributes:
      trigger_counter (int): Incremental identifier of the occurred trigger.
      trigger_mask (int): The trigger type(s) satisfied.
      overlays (int): Number of merged events.
      n_triggered_hits (int): Number of hits satisfying the trigger conditions.
      n_snapshot_hits (int): Number of snapshot hits.
      triggered_hits (list): A list of triggered hits
        (dom_id, pmt_id, tdc_time, tot)
      snapshot_hits (list): A list of snapshot hits
        (dom_id, pmt_id, tdc_time, tot)

    """
    def __init__(self, file_obj):
        self.header = DAQHeader(file_obj=file_obj)
        self.trigger_counter = unpack('<Q', file_obj.read(8))[0]
        self.trigger_mask = unpack('<Q', file_obj.read(8))[0]
        self.overlays = unpack('<i', file_obj.read(4))[0]

        self.n_triggered_hits = unpack('<i', file_obj.read(4))[0]
        self.triggered_hits = []
        self._parse_triggered_hits(file_obj)

        self.n_snapshot_hits = unpack('<i', file_obj.read(4))[0]
        self.snapshot_hits = []
        self._parse_snapshot_hits(file_obj)

    def _parse_triggered_hits(self, file_obj):
        """Parse and store triggered hits."""
        for _ in range(self.n_triggered_hits):
            dom_id, pmt_id, tdc_time, tot = unpack('<ibib', file_obj.read(10))
            self.triggered_hits.append((dom_id, pmt_id, tdc_time, tot))

    def _parse_snapshot_hits(self, file_obj):
        """Parse and store snapshot hits."""
        for _ in range(self.n_snapshot_hits):
            dom_id, pmt_id, tdc_time, tot = unpack('<ibib', file_obj.read(10))
            self.snapshot_hits.append((dom_id, pmt_id, tdc_time, tot))

    def __repr__(self):
        string = '\n'.join((
            " Number of triggered hits: " + str(self.n_triggered_hits),
            " Number of snapshot hits: " + str(self.n_snapshot_hits)
        ))
        string += "\nTriggered hits:\n"
        string += pprint.pformat(self.triggered_hits)
        string += "\nSnapshot hits:\n"
        string += pprint.pformat(self.snapshot_hits)
        return string



# coding=utf-8
# Filename: tools.py
# pylint: disable=C0103
"""
Some frequently used logic.

"""
from __future__ import division, absolute_import, print_function

__author__ = 'tamasgal'

import collections
from collections import namedtuple
from itertools import chain
import time

import numpy as np


def unpack_nfirst(seq, nfirst):
    """Unpack the nfrist items from the list and return the rest.

    >>> a, b, c, rest = unpack_nfirst((1, 2, 3, 4, 5), 3)
    >>> a, b, c
    (1, 2, 3)
    >>> rest
    (4, 5)

    """
    iterator = iter(seq)
    for _ in range(nfirst):
        yield next(iterator, None)
    yield tuple(iterator)


def split(string, callback=None):
    """Split the string and execute the callback function on each part.

    >>> string = "1 2 3 4"
    >>> parts = split(string, int)
    >>> parts
    [1, 2, 3, 4]

    """
    if callback:
        return [callback(i) for i in string.split()]
    else:
        return string.split()

def namedtuple_with_defaults(typename, field_names, default_values=[]):
    """Create a namedtuple with default values

    >>> Node = namedtuple_with_defaults('Node', 'val left right')
    >>> Node()
    Node(val=None, left=None, right=None)
    >>> Node = namedtuple_with_defaults('Node', 'val left right', [1, 2, 3])
    >>> Node()
    Node(val=1, left=2, right=3)
    >>> Node = namedtuple_with_defaults('Node', 'val left right', {'right':7})
    >>> Node()
    Node(val=None, left=None, right=7)
    >>> Node(4)
    Node(val=4, left=None, right=7)
    """
    the_tuple = namedtuple(typename, field_names)
    the_tuple.__new__.__defaults__ = (None,) * len(the_tuple._fields)
    if isinstance(default_values, collections.Mapping):
        prototype = the_tuple(**default_values)
    else:
        prototype = the_tuple(*default_values)
    the_tuple.__new__.__defaults__ = tuple(prototype)
    return the_tuple


def angle_between(v1, v2):
    """Returns the angle in radians between vectors 'v1' and 'v2'::

    >>> angle_between((1, 0, 0), (0, 1, 0))
    1.5707963267948966
    >>> angle_between((1, 0, 0), (1, 0, 0))
    0.0
    >>> angle_between((1, 0, 0), (-1, 0, 0))
    3.141592653589793

    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    angle = np.arccos(np.dot(v1_u, v2_u))
    return angle


def unit_vector(vector):
    """ Returns the unit vector of the vector.  """
    return vector / np.linalg.norm(vector)


def circ_permutation(items):
    """Calculate the circular permutation for a given list of items."""
    permutations = []
    for i in range(len(items)):
        permutations.append(items[i:] + items[:i])
    return permutations


def geant2pdg(geant_code):
    """Convert GEANT particle ID to PDG"""
    conversion_table = {
        1: 22,   # photon
        2: -11, # positron
        3: 11,   # electron
        5: -13,    # muplus
        6: 13,   # muminus
        7: 111, # pi0
        8: 211, # piplus
        9: -211, # piminus
        10: 130, # k0long
        11: 321, # kplus
        12: -321, # kminus
        13: 2112,  # neutron
        14: 2212, # proton
        16: 310, # kaon0short
        17: 221, # eta
        }
    try:
        return conversion_table[geant_code]
    except KeyError:
        return 0

def pdg2name(pdg_id):
    """Convert PDG ID to human readable names"""
    # pylint: disable=C0330
    conversion_table = {
         11: 'e-',
        -11: 'e+',
         12: 'nu_e',
        -12: 'anu_e',
         13: 'mu-',
        -13: 'mu+',
         14: 'nu_mu',
        -14: 'anu_mu',
         15: 'tau-',
        -15: 'tau+',
         16: 'nu_tau',
        -16: 'anu_tau',
         22: 'photon',
        111: 'pi0',
        130: 'K0L',
        211: 'pi-',
       -211: 'pi+',
        310: 'K0S',
        311: 'K0',
        321: 'K+',
       -321: 'K-',
       2112: 'n',
       2212: 'p',
      -2212: 'p-',
    }
    try:
        return conversion_table[pdg_id]
    except KeyError:
        return "N/A"


class PMTReplugger(object):
    """Replugs PMTs and modifies the data according to the new setup."""

    def __init__(self, pmt_combs, angles, rates):
        self._pmt_combs = pmt_combs
        self._new_combs = []
        self._angles = angles
        self._rates = rates
        self._switch = None

    def angle_for(self, pmt_comb):
        """Return angle for given PMT combination"""
        combs = self.current_combs()
        print(combs)
        idx = combs.index(pmt_comb)
        return self._angles[idx]

    def current_combs(self):
        combs = self._new_combs if self._new_combs else self._pmt_combs
        return combs

    @property
    def angles(self):
        combs = self.current_combs()
        idxs = []
        for comb in self._pmt_combs:
            idxs.append(combs.index(comb))
        angles = []
        for idx in idxs:
            angles.append(self._angles[idx])
        return angles

    def switch(self, idxs1, idxs2):
        """Switch PMTs"""
        flat_combs = np.array(self.flatten(self._pmt_combs))
        operations = []
        for old, new in zip(idxs1, idxs2):
            operations.append((self.indices(old, flat_combs), new))
        for idxs, new_value in operations:
            flat_combs[idxs] = new_value
        it = iter(flat_combs)
        self._new_combs = []
        for pmt1, pmt2 in zip(it, it):
            if pmt1 > pmt2:
                self._new_combs.append((pmt2, pmt1))
            else:
                self._new_combs.append((pmt1, pmt2))

    def reset_switches(self):
        """Reset all switches"""
        self._new_combs = None

    def indices(self, item, items):
        values = np.array(items)
        indices = np.where(values == item)[0]
        return indices

    def flatten(self, items):
        return list(chain.from_iterable(items))


class Timer(object):
    """A very simple, accurate and easy to use timer context"""
    def __enter__(self):
        self.__start = time.time()

    def __exit__(self, type, value, traceback):
        self.__finish = time.time()

    def get_seconds(self):
        return self.__finish - self.__start

    def log(self):
        print("{0}s".format(self.get_seconds()))



__author__ = 'tamasgal'


class UI(object):
    """Represents the settings for the UI."""
    fg = 'light gray'
    bg = 'black'

    palette = [
        ('default', fg, bg),
        ('highlight', fg+',standout', bg),
        ('header', 'white', 'dark cyan'),
        ('footer', 'light gray', 'dark blue'),
        ('body', 'dark cyan', '', 'standout'),
        ('focus', 'dark red', '', 'standout'),
        ('head', 'light red', 'black'),
        ('blob', 'yellow', 'dark cyan'),
        ('blob_selected', 'dark cyan', 'yellow'),
        ('blob_scale', 'dark cyan', 'black'),
        ]

    keys = {
        'select': ('return', 'enter'),
        'inspect': ('x', 'X'),
        'escape': ('esc', 'q', 'Q'),
        'left': ('left', 'h'),
        'right': ('right', 'l'),
        'up': ('up', 'k'),
        'down': ('down', 'j'),
        'home': ('0', '^'),
        'end': ('$',),
        'goto': ('g', 'G'),
        'help': ('?',),
        }



__author__ = 'tamasgal'

import urwid

from pipeinspector.widgets import BlobWidget, BlobBrowser
from pipeinspector.settings import UI






class MainFrame(urwid.Frame):
    """
    Represents the main GUI

    """
    def __init__(self, pump):
        self.header = urwid.AttrWrap(urwid.Text("PipeInspector", align='center'),
                                     'header')

        self.blob_browser = BlobBrowser()

        self.info_area = urwid.Text('')
        self.blobs = BlobWidget()
        self.footer = urwid.Columns([self.info_area, self.blobs])

        self.frame = urwid.AttrWrap(urwid.Frame(self.blob_browser,
                                                header=self.header,
                                                footer=self.footer), 'default')
        urwid.Frame.__init__(self, self.frame)
        self.overlay = None

        self.pump = pump

        urwid.connect_signal(self.blobs, 'blob_selected', self.blob_selected)
        self.blobs.goto_blob(0)

    def blob_selected(self, index):
        self.info_area.set_text("Blob: {0}".format(index))

        #blob = self.pump.process(None)
        blob = self.pump.get_blob(index)
        self.blob_browser.load(blob)

    def keypress(self, size, key):
        input = urwid.Frame.keypress(self, size, key)
        if input is None:
            return
        if input in UI.keys['left']:
            self.blobs.previous_blob()
        elif input in UI.keys['right']:
            self.blobs.next_blob()
        elif input in [key.upper() for key in UI.keys['left']]:
            self.blobs.previous_blob(step=10)
        elif input in [key.upper() for key in UI.keys['right']]:
            self.blobs.next_blob(step=10)
        elif input in UI.keys['home']:
            self.blobs.goto_blob(0)
        else:
            return self.body.keypress(size, input)



# coding=utf-8
# Filename: test_dataclasses.py
# pylint: disable=C0111,R0904,C0103
"""
...

"""
from __future__ import division, absolute_import, print_function

import numpy as np

from km3pipe.testing import TestCase
from km3pipe.dataclasses import Position, Direction


class TestPosition(TestCase):

    def test_position(self):
        position = Position((1, 2, 3))
        self.assertEqual(1, position.x)
        self.assertEqual(2, position.y)
        self.assertEqual(3, position.z)

    def test_attributes_can_be_changed(self):
        position = Position()
        position.x = 1
        position.y = 2
        position.z = 3
        self.assertEqual(1, position.x)
        self.assertEqual(1, position[0])
        self.assertEqual(2, position.y)
        self.assertEqual(2, position[1])
        self.assertEqual(3, position.z)
        self.assertEqual(3, position[2])

    def test_position_is_ndarray_like(self):
        pos = Position((1, 2, 3))
        pos *= 2
        self.assertEqual(4, pos[1])
        self.assertEqual(3, pos.size)
        self.assertTupleEqual((3,), pos.shape)


class TestDirection(TestCase):

    def test_direction(self):
        direction = Direction((1, 0, 0))
        self.assertAlmostEqual(1, np.linalg.norm(direction))
        self.assertEqual(1, direction.x)
        self.assertEqual(0, direction.y)
        self.assertEqual(0, direction.z)

    def test_direction_normalises_on_init(self):
        direction = Direction((1, 2, 3))
        self.assertAlmostEqual(0.26726124, direction.x)
        self.assertAlmostEqual(0.53452248, direction.y)
        self.assertAlmostEqual(0.80178372, direction.z)

    def test_direction_normalises_on_change_attribute(self):
        direction = Direction((1, 2, 3))
        self.assertAlmostEqual(1, np.linalg.norm(direction))
        direction.x = 10
        self.assertAlmostEqual(1, np.linalg.norm(direction))
        direction.y = 20
        self.assertAlmostEqual(1, np.linalg.norm(direction))
        direction.z = 30
        self.assertAlmostEqual(1, np.linalg.norm(direction))

    def test_direction_zenith(self):
        direction = Direction((0, 0, -1))
        self.assertAlmostEqual(0, direction.zenith)
        direction = Direction((0, 0, 1))
        self.assertAlmostEqual(np.pi, direction.zenith)
        direction = Direction((0, 1, 0))
        self.assertAlmostEqual(np.pi/2, direction.zenith)

    def test_direction_str(self):
        direction = Direction((1, 2, 3))
        self.assertEqual("(0.2673, 0.5345, 0.8018)", str(direction))





# coding=utf-8
# Filename: clb.py
"""
Pumps for the CLB data formats.

"""
from __future__ import division, absolute_import, print_function

import struct
from struct import unpack
from binascii import hexlify
from collections import namedtuple
import datetime

from km3pipe import Pump
from km3pipe.logger import logging

log = logging.getLogger(__name__)  # pylint: disable=C0103


class CLBPump(Pump):
    """A pump for binary CLB files."""

    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)
        self.filename = self.get('filename')
        self.cache_enabled = self.get('cache_enabled') or False
        self.packet_positions = []
        self.index = 0

        if self.filename:
            self.open_file(self.filename)
            if self.cache_enabled:
                self.determine_packet_positions()
        else:
            log.warn("No filename specified. Take care of the file handling!")

    def determine_packet_positions(self):
        """Record the file pointer position of each frame"""
        print("Analysing file...")
        self.rewind_file()
        try:
            while True:
                pointer_position = self.blob_file.tell()
                length = struct.unpack('<i', self.blob_file.read(4))[0]
                self.packet_positions.append(pointer_position)
                self.blob_file.seek(length, 1)
        except struct.error:
            pass
        self.rewind_file()
        print("Found {0} CLB UDP packets.".format(len(self.packet_positions)))

    def seek_to_packet(self, index):
        """Move file pointer to the packet with given index."""
        pointer_position = self.packet_positions[index]
        self.blob_file.seek(pointer_position, 0)

    def next_blob(self):
        """Generate next blob in file"""
        try:
            length = struct.unpack('<i', self.blob_file.read(4))[0]
        except struct.error:
            raise StopIteration
        header = CLBHeader(file_obj=self.blob_file)
        blob = {'CLBHeader': header}
        remaining_length = length - header.size
        pmt_data = []
        for _ in range(int(remaining_length/6)):
            channel_id, timestamp, tot = struct.unpack('>cic',
                                                       self.blob_file.read(6))
            pmt_data.append(PMTData(ord(channel_id), timestamp, ord(tot)))
        blob['PMTData'] = pmt_data
        return blob

    def get_blob(self, index):
        """Return blob at given index."""
        self.seek_to_packet(index)
        return self.next_blob()

    def process(self, blob):
        blob = self.next_blob()
        return blob

    def __iter__(self):
        return self

    def next(self):
        """Python 2/3 compatibility for iterators"""
        return self.__next__()

    def __next__(self):
        return self.next_blob()

    def finish(self):
        """Clean everything up"""
        self.blob_file.close()

class CLBHeader(object):
    """Wrapper for the CLB Common Header binary format.

    Args:
      file_obj (file): The binary file, where the file pointer is at the
        beginning of the header.

    Attributes:
      size (int): The size of the original DAQ byte representation.

    """
    size = 28

    def __init__(self, byte_data=None, file_obj=None):
        self.data_type = None
        self.run_number = None
        self.udp_sequence = None
        self.timestamp = None
        self.ns_ticks = None
        self.human_readable_timestamp = None
        self.dom_id = None
        self.dom_status = None
        self.time_valid = None
        if byte_data:
            self._parse_byte_data(byte_data)
        if file_obj:
            self._parse_file(file_obj)

    def __str__(self):
        # pylint: disable=E1124
        description = ("CLBHeader\n"
                       "    Data type:    {self.data_type}\n"
                       "    Run number:   {self.run_number}\n"
                       "    UDP sequence: {self.udp_sequence}\n"
                       "    Time stamp:   {self.timestamp}\n"
                       "                  {self.human_readable_timestamp}\n"
                       "    Ticks [16ns]: {self.ns_ticks}\n"
                       "    DOM ID:       {self.dom_id}\n"
                       "    DOM status:   {self.dom_status}\n"
                       "".format(self=self))
        return description

    def __insp__(self):
        return self.__str__()

    def _parse_byte_data(self, byte_data):
        """Extract the values from byte string."""
        self.data_type = b''.join(unpack('cccc', byte_data[:4])).decode()
        self.run_number = unpack('>i', byte_data[4:8])[0]
        self.udp_sequence = unpack('>i', byte_data[8:12])[0]
        self.timestamp, self.ns_ticks = unpack('>II', byte_data[12:20])
        self.dom_id = hexlify(b''.join(unpack('cccc',
                                              byte_data[20:24]))).decode()

        dom_status_bits = unpack('>I', byte_data[24:28])[0]
        self.dom_status = "{0:032b}".format(dom_status_bits)

        self.human_readable_timestamp = datetime.datetime.fromtimestamp(
            int(self.timestamp)).strftime('%Y-%m-%d %H:%M:%S')


    def _parse_file(self, file_obj):
        """Directly read from file handler.

        Note:
          This will move the file pointer.

        """
        byte_data = file_obj.read(self.size)
        self._parse_byte_data(byte_data)

PMTData = namedtuple('PMTData', 'channel_id timestamp tot')



# coding=utf-8
# Filename: evt.py
# pylint: disable=C0103,R0903
"""
Pumps for the EVT simulation dataformat.

"""
from __future__ import division, absolute_import, print_function

__author__ = 'tamasgal'

import sys

from collections import namedtuple

from km3pipe import Pump
from km3pipe.logger import logging

from km3pipe.dataclasses import Point, Direction
from km3pipe.tools import pdg2name, geant2pdg, unpack_nfirst

log = logging.getLogger(__name__)  # pylint: disable=C0103


class EvtPump(Pump): # pylint: disable:R0902
    """Provides a pump for EVT-files"""

    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)
        self.filename = self.get('filename')
        self.cache_enabled = self.get('cache_enabled') or False
        self.basename = self.get('basename') or None
        self.index_start = self.get('index_start') or 1
        self.index_stop = self.get('index_stop') or 1

        self.raw_header = None
        self.event_offsets = []
        self.index = 0
        self.whole_file_cached = False

        self.file_index = int(self.index_start)

        if self.basename:
            self.filename = self.basename + str(self.index_start) + '.evt'

        if self.filename:
            print("Opening {0}".format(self.filename))
            self.open_file(self.filename)
            self.prepare_blobs()
        else:
            log.warn("No filename specified. Take care of the file handling!")

    def _reset(self):
        """Clear the cache."""
        self.raw_header = None
        self.event_offsets = []
        self.index = 0

    def prepare_blobs(self):
        """Populate the blobs"""
        self.raw_header = self.extract_header()
        if self.cache_enabled:
            self._cache_offsets()

    def extract_header(self):
        """Create a dictionary with the EVT header information"""
        raw_header = self.raw_header = {}
        #for line in self.blob_file:
        first_line = self.blob_file.readline()
        self.blob_file.seek(0, 0)
        if not first_line.startswith('start_run'):
            log.warning("No header found.")
            return raw_header
        for line in iter(self.blob_file.readline, ''):
            line = line.strip()
            try:
                tag, value = line.split(':')
            except ValueError:
                continue
            raw_header[tag] = value.split()
            if line.startswith('end_event:'):
                self._record_offset()
                return raw_header
        raise ValueError("Incomplete header, no 'end_event' tag found!")

    def get_blob(self, index):
        """Return a blob with the event at the given index"""
        if index > len(self.event_offsets) - 1:
            self._cache_offsets(index, verbose=False)
        self.blob_file.seek(self.event_offsets[index], 0)
        blob = self._create_blob()
        if blob is None:
            raise IndexError
        else:
            return blob

    def process(self, blob=None):
        """Pump the next blob to the modules"""
        try:
            blob = self.get_blob(self.index)
        except IndexError:
            if self.basename and self.file_index < self.index_stop:
                self.file_index += 1
                self._reset()
                self.blob_file.close()
                self.index = 0
                self.filename = self.basename + str(self.file_index) + '.evt'
                print("Opening {0}".format(self.filename))
                self.open_file(self.filename)
                self.prepare_blobs()
                return blob
            raise StopIteration
        self.index += 1
        return blob

    def _cache_offsets(self, up_to_index=None, verbose=True):
        """Cache all event offsets."""
        if not up_to_index:
            if verbose:
                print("Caching event file offsets, this may take a minute.")
            self.blob_file.seek(0, 0)
            self.event_offsets = []
            if not self.raw_header:
                self.event_offsets.append(0)
        else:
            self.blob_file.seek(self.event_offsets[-1], 0)
        for line in iter(self.blob_file.readline, ''):
            line = line.strip()
            if line.startswith('end_event:'):
                self._record_offset()
                if len(self.event_offsets) % 100 == 0:
                    if verbose:
                        print('.', end='')
                    sys.stdout.flush()
            if up_to_index and len(self.event_offsets) >= up_to_index + 1:
                return
        self.event_offsets.pop()  # get rid of the last entry
        #self.blob_file.seek(self.event_offsets[0], 0)
        if not up_to_index:
            self.whole_file_cached = True
        print("\n{0} events indexed.".format(len(self.event_offsets)))

    def _record_offset(self):
        """Stores the current file pointer position"""
        offset = self.blob_file.tell()
        self.event_offsets.append(offset)

    def _create_blob(self):
        """Parse the next event from the current file position"""
        blob = None
        for line in self.blob_file:
            line = line.strip()
            if line.startswith('end_event:') and blob:
                blob['raw_header'] = self.raw_header
                return blob
            if line.startswith('start_event:'):
                blob = {}
                tag, value = line.split(':')
                blob[tag] = value.split()
                continue
            if blob:
                self._create_blob_entry_for_line(line, blob)

    def _create_blob_entry_for_line(self, line, blob):
        """Create the actual blob entry from the given line."""
        try:
            tag, value = line.split(':')
        except ValueError:
            log.warning("Corrupt line in EVT file:\n{0}".format(line))
            return
        if tag in ('track_in', 'track_fit', 'hit', 'hit_raw'):
            values = [float(x) for x in value.split()]
            blob.setdefault(tag, []).append(values)
            if tag == 'hit':
                blob.setdefault("EvtHits", []).append(Hit(*values))
            if tag == "hit_raw":
                blob.setdefault("EvtRawHits", []).append(RawHit(*values))
            if tag == "track_in":
                blob.setdefault("TrackIns", []).append(TrackIn(values))
            if tag == "track_fit":
                blob.setdefault("TrackFits", []).append(TrackFit(values))
        else:
            if tag == 'neutrino':
                values = [float(x) for x in value.split()]
                blob['Neutrino'] = Neutrino(values)
            else:
                blob[tag] = value.split()

    def __len__(self):
        if not self.whole_file_cached:
            self._cache_offsets()
        return len(self.event_offsets)

    def __iter__(self):
        return self

    def next(self):
        """Python 2/3 compatibility for iterators"""
        return self.__next__()

    def __next__(self):
        try:
            blob = self.get_blob(self.index)
        except IndexError:
            self.index = 0
            raise StopIteration
        self.index += 1
        return blob

    def __getitem__(self,index):
        if isinstance(index, int):
            return self.get_blob(index)
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield self.get_blob(i)

    def finish(self):
        """Clean everything up"""
        self.blob_file.close()


class Track(object):
    """Bass class for particle or shower tracks"""
#    def __init__(self, id, x, y, z, dx, dy, dz, E=None, t=0, *args):
    def __init__(self, data, zed_correction=405.93):
        id, x, y, z, dx, dy, dz, E, t, args = unpack_nfirst(data, 9)
        self.id = int(id)
        # z correctio due to gen/km3 (ZED -> sea level shift)
        # http://wiki.km3net.physik.uni-erlangen.de/index.php/Simulations
        self.pos = Point((x, y, z + zed_correction))
        self.dir = Direction((dx, dy, dz))
        self.E = E
        self.time = t
        self.args = args

    def __repr__(self):
        text = "Track:\n"
        text += " id: {0}\n".format(self.id)
        text += " pos: {0}\n".format(self.pos)
        text += " dir: {0}\n".format(self.dir)
        text += " energy: {0} GeV\n".format(self.E)
        text += " time: {0} ns\n".format(self.time)
        return text


class TrackIn(Track):
    """Representation of a track_in entry in an EVT file"""
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
        self.particle_type = geant2pdg(int(self.args[0]))
        try:
            self.length = self.args[1]
        except IndexError:
            self.length = 0

    def __repr__(self):
        text = super(self.__class__, self).__repr__()
        text += " type: {0} '{1}' [PDG]\n".format(self.particle_type,
                                                  pdg2name(self.particle_type))
        text += " length: {0} [m]\n".format(self.length)
        return text


class TrackFit(Track):
    """Representation of a track_fit entry in an EVT file"""
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
        self.speed = self.args[0]
        self.ts = self.args[1]
        self.te = self.args[2]
        self.con1 = self.args[3]
        self.con2 = self.args[4]

    def __repr__(self):
        text = super(self.__class__, self).__repr__()
        text += " speed: {0} [m/ns]\n".format(self.speed)
        text += " ts: {0} [ns]\n".format(self.ts)
        text += " te: {0} [ns]\n".format(self.te)
        text += " con1: {0}\n".format(self.con1)
        text += " con2: {0}\n".format(self.con2)
        return text


class Neutrino(object): # pylint: disable:R0902
    """Representation of a neutrino entry in an EVT file"""
    def __init__(self, data, zed_correction=405.93):
        id, x, y, z, dx, dy, dz, E, t, Bx, By, \
            ichan, particle_type, channel, args = unpack_nfirst(data, 14)
        self.id = id
        # z correctio due to gen/km3 (ZED -> sea level shift)
        # http://wiki.km3net.physik.uni-erlangen.de/index.php/Simulations
        self.pos = Point((x, y, z + zed_correction))
        self.dir = Direction((dx, dy, dz))
        self.E = E
        self.time = t
        self.Bx = Bx
        self.By = By
        self.ichan = ichan
        self.particle_type = particle_type
        self.channel = channel

    def __str__(self):
        text = "Neutrino: "
        text += pdg2name(self.particle_type)
        if self.E >= 1000000:
            text += ", {0:.3} PeV".format(self.E / 1000000)
        elif self.E >= 1000:
            text += ", {0:.3} TeV".format(self.E / 1000)
        else:
            text += ", {0:.3} GeV".format(float(self.E))
        text += ', CC' if int(self.channel) == 2 else ', NC'
        return text


# The hit entry in an EVT file
Hit = namedtuple('Hit', 'id pmt_id pe time type n_photons track_in c_time')
Hit.__new__.__defaults__ = (None, None, None, None, None, None, None, None)


# The hit_raw entry in an EVT file
def __add_raw_hit__(self, other):
    """Add two hits by adding the ToT and preserve time and pmt_id
    of the earlier one."""
    first = self if self.time <= other.time else other
    return RawHit(first.id, first.pmt_id, self.tot+other.tot, first.time)
RawHit = namedtuple('RawHit', 'id pmt_id tot time')
RawHit.__new__.__defaults__ = (None, None, None, None)
RawHit.__add__ = __add_raw_hit__




# coding=utf-8
# Filename: test_tools.py
# pylint: disable=locally-disabled,C0111,R0904,C0103
from __future__ import division, absolute_import, print_function

import numpy as np
import itertools

from km3pipe.testing import TestCase
from km3pipe.tools import (unpack_nfirst, split, namedtuple_with_defaults,
                           angle_between, geant2pdg, pdg2name, PMTReplugger)


class TestTools(TestCase):

    def test_unpack_nfirst(self):
        a_tuple = (1, 2, 3, 4, 5)
        a, b, c, rest = unpack_nfirst(a_tuple, 3)
        self.assertEqual(1, a)
        self.assertEqual(2, b)
        self.assertEqual(3, c)
        self.assertTupleEqual((4, 5), rest)

    def test_split_splits_strings(self):
        string = "1 2 3 4"
        parts = split(string)
        self.assertListEqual(['1', '2', '3', '4'], parts)

    def test_split_callback_converts_correctly(self):
        string = "1 2 3 4"
        parts = split(string, int)
        self.assertListEqual([1, 2, 3, 4], parts)

        string = "1.0 2.1 3.2 4.3"
        parts = split(string, float)
        self.assertListEqual([1.0, 2.1, 3.2, 4.3], parts)

    def test_namedtuple_with_defaults_initialises_with_none(self):
        Node = namedtuple_with_defaults('Node', 'val left right')
        node = Node()
        self.assertIsNone(node.val)
        self.assertIsNone(node.left)
        self.assertIsNone(node.right)

    def test_namedtuple_with_defaults_initialises_with_given_values(self):
        Node = namedtuple_with_defaults('Node', 'val left right', [1, 2, 3])
        node = Node()
        self.assertEqual(1, node.val)
        self.assertEqual(2, node.left)
        self.assertEqual(3, node.right)

    def test_angle_between(self):
        v1 = (1, 0, 0)
        v2 = (0, 1, 0)
        v3 = (-1, 0, 0)
        self.assertAlmostEqual(0, angle_between(v1, v1))
        self.assertAlmostEqual(np.pi/2, angle_between(v1, v2))
        self.assertAlmostEqual(np.pi, angle_between(v1, v3))

    def test_angle_between_returns_nan_for_zero_length_vectors(self):
        v1 = (0, 0, 0)
        v2 = (1, 0, 0)
        self.assertTrue(np.isnan(angle_between(v1, v2)))

    def test_geant2pdg(self):
        self.assertEqual(22, geant2pdg(1))
        self.assertEqual(-13, geant2pdg(5))

    def test_geant2pdg_returns_0_for_unknown_particle_id(self):
        self.assertEqual(0, geant2pdg(-999))

    def test_pdg2name(self):
        self.assertEqual('mu-', pdg2name(13))
        self.assertEqual('anu_tau', pdg2name(-16))

    def test_pdg2name_returns_NA_for_unknown_particle(self):
        self.assertEqual('N/A', pdg2name(0))


# [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)]
PMT_COMBS = list(itertools.combinations(range(4), 2))
ANGLES = range(len(PMT_COMBS))

class TestPMTReplugger(TestCase):

    def setUp(self):
        self.replugger = PMTReplugger(PMT_COMBS, ANGLES, [])

    def test_angle_for(self):
        #self.assertEqual(0, self.replugger.angle_for((0, 1)))
        #self.assertEqual(1, self.replugger.angle_for((0, 2)))
        pass

    def test_switch(self):
        self.replugger.switch([0, 1], [1, 0])
        self.assertEqual(self.replugger._new_combs,
                         [(0, 1), (1, 2), (1, 3), (0, 2), (0, 3), (2, 3)])

    def test_switch_three_indicies(self):
        self.replugger.switch([0, 1, 2], [1, 2, 0])
        self.assertEqual(self.replugger._new_combs,
                         [(1, 2), (0, 1), (1, 3), (0, 2), (2, 3), (0, 3)])

    def test_angle_is_correct_if_two_pmts_are_switched(self):
        self.replugger.switch([0, 1], [1, 0])
        self.assertEqual(0, self.replugger.angle_for((0, 1)))
        self.assertEqual(3, self.replugger.angle_for((0, 2)))
        self.assertEqual(4, self.replugger.angle_for((0, 3)))

    def test_angles_are_ordered_correctly_after_switch(self):
        self.replugger.switch([0, 1, 2], [1, 2, 0])
        self.assertListEqual([1, 3, 5, 0, 2, 4], self.replugger.angles)



#!/usr/bin/env python
# coding=utf-8
# Filename: jpp.py
# pylint: disable=
"""
Pump for the jpp file read through aanet interface.

"""
from __future__ import division, absolute_import, print_function

from km3pipe import Pump
from km3pipe.logger import logging

log = logging.getLogger(__name__)  # pylint: disable=C0103


class JPPPump(Pump):
    """A pump for JPP ROOT files."""

    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)

        self.index = self.get('index') or 0

        import aa
        import ROOT
        if self.get('index'):
            self.index = self.get('index')
        else:
            self.index = 0

        self.index_start = self.get('index_start') or 1
        self.index_stop = self.get('index_stop') or 1

        self.filename = self.get('filename')
        self.basename = self.get('basename')
        if not self.filename and not self.basename:
            raise ValueError("No filename defined")

        self.file_index = self.index_start

        if self.basename:
            self.rootfile = ROOT.EventFile(self.basename + str(self.file_index) + ".JTE.root")

        else:
                self.rootfile = ROOT.EventFile(self.filename)

        self.evt = ROOT.Evt()

    def get_blob(self, index):
        """Return the blob"""
        self.rootfile.set_index(index)
        self.evt = self.rootfile.evt
        return {'Evt': self.evt}

    def process(self, blob):
        if self.rootfile.set_index(self.index):
            self.evt = self.rootfile.evt
            self.index += 1
            return {'Evt': self.evt}
        else:
            self.file_index += 1
            if self.basename and self.file_index <= self.index_stop:
                import aa
                import ROOT
                print("open next file")
                self.rootfile = ROOT.EventFile(self.basename + str(self.file_index) + ".JTE.root")
                self.index = 0
                self.process(blob)
            else:
                raise StopIteration


    def finish(self):
        self.rootfile.Close()



# coding=utf-8
# Filename: test_evt.py
# pylint: disable=locally-disabled,C0111,R0904,C0301,C0103,W0212
from __future__ import division, absolute_import, print_function

__author__ = 'tamasgal'

import operator
from functools import reduce

from km3pipe.testing import TestCase, StringIO
from km3pipe.pumps import EvtPump
from km3pipe.pumps.evt import Track, TrackIn, Neutrino, Hit, RawHit, TrackFit


class TestEvtParser(TestCase):

    def setUp(self):
        self.valid_evt_header = "\n".join((
            "start_run: 1",
            "cut_nu: 0.100E+03 0.100E+09-0.100E+01 0.100E+01",
            "spectrum: -1.40",
            "end_event:",
            "start_event: 12 1",
            "track_in: 1 -389.951 213.427 516 -0.204562 -0.60399 -0.770293 9.092 0 5 40.998",
            "hit: 1 44675 1 1170.59 5 2 1 1170.59",
            "end_event:",
            "start_event: 13 1",
            "track_in:  1 272.695 -105.613 516 -0.425451 -0.370522 -0.825654 2431.47 0 5 -1380",
            "track_in:  2 272.348 -106.292 516 -0.425451 -0.370522 -0.825654 24670.7 1.33 5 -1484",
            "track_in:  3 279.47 -134.999 516 -0.425451 -0.370522 -0.825654 164.586 26.7 5 601.939",
            "hit: 1 20140 1 1140.06 5 1 1 1140.06",
            "hit: 2 20159 1 1177.14 5 1 1 1177.14",
            "hit: 3 20164 1 1178.19 5 1 1 1178.19",
            "hit: 4 20170 1 1177.23 5 1 1 1177.23",
            "hit: 5 20171 2 1177.25 5 1 2 1177.25",
            "end_event:",
            "start_event: 14 1",
            "track_in:  1 40.256 -639.888 516 0.185998 0.476123 -0.859483 10016.1 0 5 -1668",
            "hit: 1 33788 1 2202.81 5 1 1 2202.81",
            "hit: 2 33801 1 2248.95 5 1 1 2248.95",
            "hit: 3 33814 1 2249.2 5 1 1 2249.2",
            "end_event:"
        ))
        self.no_evt_header = "\n".join((
            "start_event: 12 1",
            "track_in: 1 -389.951 213.427 516 -0.204562 -0.60399 -0.770293 9.092 0 5 40.998",
            "hit: 1 44675 1 1170.59 5 2 1 1170.59",
            "end_event:",
            "start_event: 13 1",
            "track_in:  1 272.695 -105.613 516 -0.425451 -0.370522 -0.825654 2431.47 0 5 -1380",
            "track_in:  2 272.348 -106.292 516 -0.425451 -0.370522 -0.825654 24670.7 1.33 5 -1484",
            "track_in:  3 279.47 -134.999 516 -0.425451 -0.370522 -0.825654 164.586 26.7 5 601.939",
            "hit: 1 20140 1 1140.06 5 1 1 1140.06",
            "hit: 2 20159 1 1177.14 5 1 1 1177.14",
            "hit: 3 20164 1 1178.19 5 1 1 1178.19",
            "hit: 4 20170 1 1177.23 5 1 1 1177.23",
            "hit: 5 20171 2 1177.25 5 1 2 1177.25",
            "end_event:",
            "start_event: 14 1",
            "track_in:  1 40.256 -639.888 516 0.185998 0.476123 -0.859483 10016.1 0 5 -1668",
            "hit: 1 33788 1 2202.81 5 1 1 2202.81",
            "hit: 2 33801 1 2248.95 5 1 1 2248.95",
            "hit: 3 33814 1 2249.2 5 1 1 2249.2",
            "end_event:"
        ))
        self.corrupt_evt_header = "foo"
        self.corrupt_line = "\n".join((
            "start_event: 1 1",
            "corrupt line",
            "end_event:"
            ))

        self.pump = EvtPump()
        self.pump.blob_file = StringIO(self.valid_evt_header)

    def tearDown(self):
        self.pump.blob_file.close()

    def test_parse_header(self):
        raw_header = self.pump.extract_header()
        self.assertEqual(['1'], raw_header['start_run'])
        self.assertAlmostEqual(-1.4, float(raw_header['spectrum'][0]))
        self.assertAlmostEqual(1, float(raw_header['cut_nu'][2]))

#    def test_incomplete_header_raises_value_error(self):
#        temp_file = StringIO(self.corrupt_evt_header)
#        pump = EvtPump()
#        pump.blob_file = temp_file
#        with self.assertRaises(ValueError):
#            pump.extract_header()
#        temp_file.close()

    def test_record_offset_saves_correct_offset(self):
        self.pump.blob_file = StringIO('a'*42)
        offsets = [1, 4, 9, 12, 23]
        for offset in offsets:
            self.pump.blob_file.seek(0, 0)
            self.pump.blob_file.seek(offset, 0)
            self.pump._record_offset()
        self.assertListEqual(offsets, self.pump.event_offsets)

    def test_event_offset_is_at_first_event_after_parsing_header(self):
        self.pump.extract_header()
        self.assertEqual(88, self.pump.event_offsets[0])

    def test_rebuild_offsets(self):
        self.pump.extract_header()
        self.pump._cache_offsets()
        self.assertListEqual([88, 233, 700], self.pump.event_offsets)

    def test_rebuild_offsets_without_header(self):
        self.pump.blob_file = StringIO(self.no_evt_header)
        self.pump.extract_header()
        self.pump._cache_offsets()
        self.assertListEqual([0, 145, 612], self.pump.event_offsets)

    def test_cache_enabled_triggers_rebuild_offsets(self):
        self.pump.cache_enabled = True
        self.pump.prepare_blobs()
        self.assertEqual(3, len(self.pump.event_offsets))

    def test_cache_disabled_doesnt_trigger_cache_offsets(self):
        self.pump.cache_enabled = False
        self.pump.prepare_blobs()
        self.assertEqual(1, len(self.pump.event_offsets))

    def test_get_blob_triggers_cache_offsets_if_cache_disabled_and_asking_for_not_indexed_event(self):
        self.pump.cache_enabled = False
        self.pump.prepare_blobs()
        self.assertEqual(1, len(self.pump.event_offsets))
        blob = self.pump.get_blob(2)
        self.assertListEqual(['14', '1'], blob['start_event'])
        self.assertEqual(3, len(self.pump.event_offsets))

    def test_get_blob_raises_index_error_for_wrong_index(self):
        self.pump.prepare_blobs()
        with self.assertRaises(IndexError):
            self.pump.get_blob(23)

    def test_get_blob_returns_correct_event_information(self):
        self.pump.prepare_blobs()
        blob = self.pump.get_blob(0)
        self.assertTrue('raw_header' in blob)
        self.assertEqual(['1'], blob['raw_header']['start_run'])
        self.assertListEqual(['12', '1'], blob['start_event'])
        self.assertListEqual([[1.0, 44675.0, 1.0, 1170.59, 5.0, 2.0, 1.0, 1170.59]],
                             blob['hit'])

    def test_get_blob_returns_correct_events(self):
        self.pump.prepare_blobs()
        blob = self.pump.get_blob(0)
        self.assertListEqual(['12', '1'], blob['start_event'])
        blob = self.pump.get_blob(2)
        self.assertListEqual(['14', '1'], blob['start_event'])
        blob = self.pump.get_blob(1)
        self.assertListEqual(['13', '1'], blob['start_event'])

    def test_process_returns_correct_blobs(self):
        self.pump.prepare_blobs()
        blob = self.pump.process()
        self.assertListEqual(['12', '1'], blob['start_event'])
        blob = self.pump.process()
        self.assertListEqual(['13', '1'], blob['start_event'])
        blob = self.pump.process()
        self.assertListEqual(['14', '1'], blob['start_event'])

    def test_process_raises_stop_iteration_if_eof_reached(self):
        self.pump.prepare_blobs()
        self.pump.process()
        self.pump.process()
        self.pump.process()
        with self.assertRaises(StopIteration):
            self.pump.process()

    def test_pump_acts_as_iterator(self):
        self.pump.prepare_blobs()
        event_numbers = []
        for blob in self.pump:
            event_numbers.append(blob['start_event'][0])
        self.assertListEqual(['12', '13', '14'], event_numbers)


    def test_pump_has_len(self):
        self.pump.prepare_blobs()
        self.assertEqual(3, len(self.pump))

    def test_pump_get_item_returns_first_for_index_zero(self):
        self.pump.prepare_blobs()
        first_blob = self.pump[0]
        self.assertEqual('12', first_blob['start_event'][0])

    def test_pump_get_item_returns_correct_blob_for_index(self):
        self.pump.prepare_blobs()
        blob = self.pump[1]
        self.assertEqual('13', blob['start_event'][0])

    def test_pump_slice_generator(self):
        self.pump.prepare_blobs()
        blobs = self.pump[:]
        blobs = list(self.pump[1:3])
        self.assertEqual(2, len(blobs))
        self.assertEqual(['13', '1'], blobs[0]['start_event'])

    def test_create_blob_entry_for_line_ignores_corrupt_line(self):
        self.pump.blob_file = StringIO(self.corrupt_line)
        self.pump.extract_header()
        self.pump.prepare_blobs()
        self.pump.get_blob(0)


class TestTrack(TestCase):
    def setUp(self):
        self.track = Track((1, 2, 3, 4, 0, 0, 1, 8, 9, 'a', 'b', 'c'),
                           zed_correction=0)

    def test_track_init(self):
        track = self.track
        self.assertEqual(1, track.id)
        self.assertListEqual([2, 3, 4], list(track.pos))
        self.assertListEqual([0, 0, 1], list(track.dir))
        self.assertEqual(8, track.E)
        self.assertEqual(9, track.time)
        self.assertTupleEqual(('a', 'b', 'c'), track.args)

    def test_track_repr(self):
        repr_str = ("Track:\n id: 1\n pos: [2 3 4]\n dir: (0.0, 0.0, 1.0)\n"
                    " energy: 8 GeV\n time: 9 ns\n")
        self.assertEqual(repr_str, self.track.__repr__())


class TestTrackIn(TestCase):

    def setUp(self):
        self.track_in = TrackIn((1, 2, 3, 4, 0, 0, 1, 8, 9, 10, 11),
                                zed_correction=0)

    def test_trackin_init(self):
        track_in = self.track_in
        self.assertEqual(1, track_in.id)
        self.assertListEqual([2, 3, 4], list(track_in.pos))
        self.assertListEqual([0, 0, 1], list(track_in.dir))
        self.assertEqual(8, track_in.E)
        self.assertEqual(9, track_in.time)
        self.assertEqual(130, track_in.particle_type) # this should be PDG!
        self.assertEqual(11, track_in.length)

    def test_track_repr(self):
        repr_str = ("Track:\n id: 1\n pos: [2 3 4]\n dir: (0.0, 0.0, 1.0)\n"
                    " energy: 8 GeV\n time: 9 ns\n type: 130 'K0L' [PDG]\n"
                    " length: 11 [m]\n")
        self.assertEqual(repr_str, self.track_in.__repr__())


class TestTrackFit(TestCase):

    def setUp(self):
        data = (1, 2, 3, 4, 0, 0, 1, 8, 9, 10, 11, 12, 13, 14)
        self.track_fit = TrackFit(data, zed_correction=0)

    def test_trackfit_init(self):
        track_fit = self.track_fit
        self.assertEqual(1, track_fit.id)
        self.assertListEqual([2, 3, 4], list(track_fit.pos))
        self.assertListEqual([0, 0, 1], list(track_fit.dir))
        self.assertEqual(8, track_fit.E)
        self.assertEqual(9, track_fit.time)
        self.assertEqual(10, track_fit.speed)
        self.assertEqual(11, track_fit.ts)
        self.assertEqual(12, track_fit.te)
        self.assertEqual(13, track_fit.con1)
        self.assertEqual(14, track_fit.con2)

    def test_trackfit_repr(self):
        repr_str = ("Track:\n id: 1\n pos: [2 3 4]\n dir: (0.0, 0.0, 1.0)\n "
                    "energy: 8 GeV\n time: 9 ns\n speed: 10 [m/ns]\n"
                    " ts: 11 [ns]\n te: 12 [ns]\n con1: 13\n con2: 14\n")
        self.assertEqual(repr_str, self.track_fit.__repr__())


class TestNeutrino(TestCase):

    def setUp(self):
        data = (1, 2, 3, 4, 0, 0, 1, 8, 9, 10, 11, 12, 13, 14)
        self.neutrino = Neutrino(data, zed_correction=0)

    def test_neutrino_init(self):
        neutrino = self.neutrino
        self.assertEqual(1, neutrino.id)
        self.assertListEqual([2, 3, 4], list(neutrino.pos))
        self.assertListEqual([0, 0, 1], list(neutrino.dir))
        self.assertEqual(8, neutrino.E)
        self.assertEqual(9, neutrino.time)
        self.assertEqual(10, neutrino.Bx)
        self.assertEqual(11, neutrino.By)
        self.assertEqual(12, neutrino.ichan)
        self.assertEqual(13, neutrino.particle_type)
        self.assertEqual(14, neutrino.channel)

    def test_neutrino_str(self):
        neutrino = self.neutrino
        repr_str = "Neutrino: mu-, 8.0 GeV, NC"
        self.assertEqual(repr_str, str(neutrino))
        neutrino.E = 2000
        repr_str = "Neutrino: mu-, 2.0 TeV, NC"
        self.assertEqual(repr_str, str(neutrino))
        neutrino.E = 3000000
        repr_str = "Neutrino: mu-, 3.0 PeV, NC"
        self.assertEqual(repr_str, str(neutrino))

class TestHit(TestCase):

    def test_hit_init(self):
        hit = Hit(1, 2, 3, 4, 5, 6, 7, 8)
        self.assertEqual(1, hit.id)
        self.assertEqual(2, hit.pmt_id)
        self.assertEqual(3, hit.pe)
        self.assertEqual(4, hit.time)
        self.assertEqual(5, hit.type)
        self.assertEqual(6, hit.n_photons)
        self.assertEqual(7, hit.track_in)
        self.assertEqual(8, hit.c_time)

    def test_hit_default_values(self):
        hit = Hit()
        self.assertIsNone(hit.id)
        self.assertIsNone(hit.pmt_id)
        self.assertIsNone(hit.time)

    def test_hit_default_values_are_set_if_others_are_given(self):
        hit = Hit(track_in=1)
        self.assertIsNone(hit.id)
        self.assertIsNone(hit.time)

    def test_hit_attributes_are_immutable(self):
        hit = Hit(1, True)
        with self.assertRaises(AttributeError):
            hit.time = 10




class TestRawHit(TestCase):

    def test_rawhit_init(self):
        raw_hit = RawHit(1, 2, 3, 4)
        self.assertEqual(1, raw_hit.id)
        self.assertEqual(2, raw_hit.pmt_id)
        self.assertEqual(3, raw_hit.tot)
        self.assertEqual(4, raw_hit.time)

    def test_hit_default_values(self):
        raw_hit = RawHit()
        self.assertIsNone(raw_hit.id)
        self.assertIsNone(raw_hit.pmt_id)
        self.assertIsNone(raw_hit.time)

    def test_hit_default_values_are_set_if_others_are_given(self):
        raw_hit = RawHit(pmt_id=1)
        self.assertIsNone(raw_hit.id)
        self.assertIsNone(raw_hit.time)

    def test_hit_attributes_are_immutable(self):
        raw_hit = RawHit(1, True)
        with self.assertRaises(AttributeError):
            raw_hit.time = 10

    def test_hit_addition_remains_time_id_and_pmt_id_and_adds_tot(self):
        hit1 = RawHit(id=1, time=1, pmt_id=1, tot=10)
        hit2 = RawHit(id=2, time=2, pmt_id=2, tot=20)
        merged_hit = hit1 + hit2
        self.assertEqual(1, merged_hit.id)
        self.assertEqual(1, merged_hit.time)
        self.assertEqual(1, merged_hit.pmt_id)
        self.assertEqual(30, merged_hit.tot)

    def test_hit_addition_picks_correct_time_if_second_hit_is_the_earlier_one(self):
        hit1 = RawHit(id=1, time=2, pmt_id=1, tot=10)
        hit2 = RawHit(id=2, time=1, pmt_id=2, tot=20)
        merged_hit = hit1 + hit2
        self.assertEqual(2, merged_hit.id)
        self.assertEqual(2, merged_hit.pmt_id)

    def test_hit_additions_works_with_multiple_hits(self):
        hit1 = RawHit(id=1, time=2, pmt_id=1, tot=10)
        hit2 = RawHit(id=2, time=1, pmt_id=2, tot=20)
        hit3 = RawHit(id=3, time=1, pmt_id=3, tot=30)
        merged_hit = hit1 + hit2 + hit3
        self.assertEqual(2, merged_hit.pmt_id)
        self.assertEqual(60, merged_hit.tot)
        self.assertEqual(1, merged_hit.time)
        self.assertEqual(2, merged_hit.id)

    def test_hit_addition_works_with_sum(self):
        hit1 = RawHit(id=1, time=2, pmt_id=1, tot=10)
        hit2 = RawHit(id=2, time=1, pmt_id=2, tot=20)
        hit3 = RawHit(id=3, time=1, pmt_id=3, tot=30)
        hits = [hit1, hit2, hit3]
        merged_hit = reduce(operator.add, hits)
        self.assertEqual(2, merged_hit.id)
        self.assertEqual(1, merged_hit.time)
        self.assertEqual(60, merged_hit.tot)
        self.assertEqual(2, merged_hit.pmt_id)




# coding=utf-8
# Filename: test_aanet.py
# pylint: disable=C0111,R0904,R0201
"""
...

"""
from __future__ import division, absolute_import, print_function

from km3pipe.testing import *

from km3pipe.pumps.aanet import AanetPump

#try:
#    # pylint: disable=F0401,W0611
#    import aa
#except ImportError:
#    NO_AA = True
#else:
#    NO_AA = False

#import aa
#import ROOT

#@skipIf(NO_AA, "Skipping tests for aanet")
#class TestAanetPump(TestCase):

#    def test_aanetpump_init(self):
#        pump = AanetPump()




# coding=utf-8
# Filename: test_decorators.py
# pylint: disable=C0111,R0904,C0103,R0903,R0201
from __future__ import division, absolute_import, print_function

from km3pipe.testing import TestCase, StringIO
from km3pipe.decorators import remain_file_pointer


class TestRemainFilePointer(TestCase):

    def test_remains_file_pointer_in_function(self):
        dummy_file = StringIO('abcdefg')

        @remain_file_pointer
        def seek_into_file(file_obj):
            file_obj.seek(1, 0)

        dummy_file.seek(2, 0)
        self.assertEqual(2, dummy_file.tell())
        seek_into_file(dummy_file)
        self.assertEqual(2, dummy_file.tell())

    def test_remains_file_pointer_and_return_value_in_function(self):
        dummy_file = StringIO('abcdefg')

        @remain_file_pointer
        def seek_into_file(file_obj):
            file_obj.seek(1, 0)
            return 1

        dummy_file.seek(2, 0)
        self.assertEqual(2, dummy_file.tell())
        return_value = seek_into_file(dummy_file)
        self.assertEqual(2, dummy_file.tell())
        self.assertEqual(1, return_value)

    def test_remains_file_pointer_in_class_method(self):

        class FileSeekerClass(object):
            def __init__(self):
                self.dummy_file = StringIO('abcdefg')

            @remain_file_pointer
            def seek_into_file(self, file_obj):
                file_obj.seek(1, 0)

        fileseeker = FileSeekerClass()
        fileseeker.dummy_file.seek(2, 0)
        self.assertEqual(2, fileseeker.dummy_file.tell())
        fileseeker.seek_into_file(fileseeker.dummy_file)
        self.assertEqual(2, fileseeker.dummy_file.tell())

    def test_remains_file_pointer_and_return_value_in_class_method(self):

        class FileSeekerClass(object):
            def __init__(self):
                self.dummy_file = StringIO('abcdefg')

            @remain_file_pointer
            def seek_into_file(self, file_obj):
                file_obj.seek(1, 0)
                return 1

        fileseeker = FileSeekerClass()
        fileseeker.dummy_file.seek(2, 0)
        self.assertEqual(2, fileseeker.dummy_file.tell())
        return_value = fileseeker.seek_into_file(fileseeker.dummy_file)
        self.assertEqual(2, fileseeker.dummy_file.tell())
        self.assertEqual(1, return_value)



# coding=utf-8
# Filename: logger.py
# pylint: disable=locally-disabled,C0103
"""
The logging facility.

"""
from __future__ import division, absolute_import, print_function

__author__ = 'tamasgal'

import logging
import logging.config
try:
    logging.config.fileConfig('logging.conf')
except Exception:
    logging.basicConfig()

logging.addLevelName(logging.INFO, "\033[1;32m%s\033[1;0m" %
                     logging.getLevelName(logging.INFO))
logging.addLevelName(logging.DEBUG, "\033[1;34m%s\033[1;0m" %
                     logging.getLevelName(logging.DEBUG))
logging.addLevelName(logging.WARNING, "\033[1;33m%s\033[1;0m" %
                     logging.getLevelName(logging.WARNING))
logging.addLevelName(logging.ERROR, "\033[1;31m%s\033[1;0m" %
                     logging.getLevelName(logging.ERROR))

# pylint: disable=C0103
formatter = logging.Formatter('[%(levelname)s] %(name)s: %(message)s')






# coding=utf-8
# Filename: __init__.py
"""
A collection of pumps for different kinds of data formats.

"""
from __future__ import division, absolute_import, print_function

from km3pipe.pumps.evt import EvtPump
from km3pipe.pumps.daq import DAQPump
from km3pipe.pumps.clb import CLBPump
from km3pipe.pumps.aanet import AanetPump
from km3pipe.pumps.jpp import JPPPump
from km3pipe.pumps.ch import CHPump



# coding=utf-8
# Filename: __init__.py
# pylint: disable=locally-disabled
"""
A collection of commonly used modules.

"""
from __future__ import division, absolute_import, print_function

import timeit

from km3pipe import Module


class HitCounter(Module):
    """Prints the number of hits and raw hits in an Evt file"""
    def process(self, blob):
        try:
            print("Number of hits: {0}".format(len(blob['hit'])))
        except KeyError:
            pass
        try:
            print("Number of raw hits: {0}".format(len(blob['hit_raw'])))
        except KeyError:
            pass
        return blob


class BlobIndexer(Module):
    """Puts an incremented index in each blob for the key 'blob_index'"""
    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)
        self.blob_index = 0

    def process(self, blob):
        blob['blob_index'] = self.blob_index
        self.blob_index += 1
        return blob


class StatusBar(Module):
    """Displays the current blob number"""
    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)
        self.blob_index = 0
        self.start = timeit.default_timer()

    def process(self, blob):
        print("------------[Blob {0:>7}]-------------".format(self.blob_index))
        self.blob_index += 1
        return blob

    def finish(self):
        """Display some basic statistics like elapsed time"""
        elapsed_time = timeit.default_timer() - self.start
        print("\n" + '='*42)
        print("Processed {0} blobs in {1} s."
              .format(self.blob_index, elapsed_time))





# coding=utf-8
# Filename: test_daq.py
# pylint: disable=C0111,R0904,C0103
"""
...

"""
from __future__ import division, absolute_import, print_function

from km3pipe.testing import TestCase, StringIO
from km3pipe.pumps.daq import (DAQPump, DAQPreamble, DAQHeader,
                               DAQSummaryslice, DAQEvent)

from binascii import unhexlify


HEX_DATA = ''.join("85000000d1070000ae01000001000000000000000000000003000000" +
                   "65000000000000000000000000000000000000000000000000000000" +
                   "00000000000000660000000000000000000000000000000000000000" +
                   "00000000000000000000000000006700000000000000000000000000" +
                   "00000000000000000000000000000000000000000070000000112700" +
                   "00ae0100000200000000000000000080000000000000000000020000" +
                   "000000000000000000020000006700000007f5f40500306700000012" +
                   "f8f405002704000000660000000a6df50500186700000007f5f40500" +
                   "306700000012f8f4050027670000000c1bf605002166000000112700" +
                   "00ae0100000200000000000000000080000100000000000000020000" +
                   "0000000000000000000200000067000000108b1e0b00266700000011" +
                   "901e0b0012030000006500000016f41f0b002467000000108b1e0b00" +
                   "266700000011901e0b00125c00000011270000ae0100000200000000" +
                   "00000000008000020000000000000002000000000000000000000002" +
                   "00000066000000088b950f0023660000001884950f001a0200000066" +
                   "0000001884950f001a66000000088b950f00235c00000011270000ae" +
                   "01000002000000000000000000800003000000000000000200000000" +
                   "000000000000000200000065000000040a1a130022650000000a0e1a" +
                   "1300280200000065000000040a1a130022650000000a0e1a130028")
BINARY_DATA = unhexlify(HEX_DATA.encode())
try:
    TEST_FILE = StringIO(BINARY_DATA)
except TypeError:
    from io import BytesIO
    TEST_FILE = BytesIO(BINARY_DATA)


class TestDAQPump(TestCase):

    def setUp(self):
        TEST_FILE.seek(0, 0)
        self.pump = DAQPump()
        self.pump.blob_file = TEST_FILE

    def test_determine_frame_positions(self):
        pump = self.pump
        pump.determine_frame_positions()
        self.assertListEqual([0, 133, 245, 347, 439], pump.frame_positions)

    def test_next_blob_finds_correct_frame_types(self):
        pump = self.pump
        blob = pump.next_blob()
        self.assertTrue('DAQSummaryslice' in blob)
        blob = pump.next_blob()
        self.assertTrue('DAQEvent' in blob)
        blob = pump.next_blob()
        self.assertTrue('DAQEvent' in blob)
        blob = pump.next_blob()
        self.assertTrue('DAQEvent' in blob)
        blob = pump.next_blob()
        self.assertTrue('DAQEvent' in blob)

    def test_next_blob_raises_stop_iteration_when_eof_reached(self):
        pump = self.pump
        with self.assertRaises(StopIteration):
            for i in range(6):
                pump.next_blob()

    def test_seek_to_frame(self):
        pump = self.pump
        pump.determine_frame_positions()
        pump.seek_to_frame(2)
        self.assertEqual(245, pump.blob_file.tell())

    def test_next_blob_returns_correct_frame_after_seek_to_frame(self):
        pump = self.pump
        pump.determine_frame_positions()
        pump.seek_to_frame(1)
        blob = pump.next_blob()
        self.assertEqual(4, blob['DAQEvent'].n_snapshot_hits)

    def test_get_blob(self):
        pump = self.pump
        pump.determine_frame_positions()
        blob = pump.get_blob(2)
        self.assertEqual(3, blob['DAQEvent'].n_snapshot_hits)

class TestDAQPreamble(TestCase):

    def setUp(self):
        TEST_FILE.seek(0, 0)

    def test_init_with_byte_data(self):
        byte_data = unhexlify("85000000D1070000".encode())
        preamble = DAQPreamble(byte_data=byte_data)
        self.assertEqual(133, preamble.length)
        self.assertEqual(2001, preamble.data_type)

    def test_parse_from_file(self):
        self.setUp()
        preamble = DAQPreamble(file_obj=TEST_FILE)
        self.assertEqual(133, preamble.length)
        self.assertEqual(2001, preamble.data_type)



class TestDAQHeader(TestCase):

    def test_init_with_byte_data(self):
        byte_data = unhexlify("AE010000010000000000000000000000".encode())
        header = DAQHeader(byte_data=byte_data)
        self.assertEqual(430, header.run)
        self.assertEqual(1, header.time_slice)

    def test_parse_from_file(self):
        TEST_FILE.seek(8, 0)  # skip preamble
        self.setUp()
        header = DAQHeader(file_obj=TEST_FILE)
        self.assertEqual(430, header.run)
        self.assertEqual(1, header.time_slice)


class TestDAQSummaryslice(TestCase):

    def test_init_with_a_slice(self):
        TEST_FILE.seek(0, 0)
        DAQPreamble(file_obj=TEST_FILE)
        sum_slice = DAQSummaryslice(TEST_FILE)
        self.assertEqual(3, sum_slice.n_summary_frames)
        print(sum_slice.summary_frames.keys())
        self.assertListEqual([101, 102, 103],
                             list(sum_slice.summary_frames.keys()))
        self.assertEqual(31, len(sum_slice.summary_frames[101]))
        self.assertEqual(31, len(sum_slice.summary_frames[102]))
        self.assertEqual(31, len(sum_slice.summary_frames[103]))


class TestDAQEvent(TestCase):

    def setUp(self):
        TEST_FILE.seek(245, 0)
        DAQPreamble(file_obj=TEST_FILE)
        self.event = DAQEvent(TEST_FILE)

    def test_init_with_a_frame(self):
        event = self.event
        self.assertEqual(1, event.trigger_counter)
        self.assertEqual(2, event.trigger_mask)
        self.assertEqual(0, event.overlays)
        self.assertEqual(2, event.n_triggered_hits)

    def test_triggered_hits(self):
        event = self.event
        self.assertEqual(2, len(event.triggered_hits))
        self.assertTupleEqual((103, 16, 728715, 38), event.triggered_hits[0])
        self.assertTupleEqual((103, 17, 728720, 18), event.triggered_hits[1])

    def test_snapshot_hits(self):
        event = self.event
        self.assertEqual(3, event.n_snapshot_hits)
        self.assertEqual(3, len(event.snapshot_hits))
        self.assertTupleEqual((101, 22, 729076, 36), event.snapshot_hits[0])
        self.assertTupleEqual((103, 16, 728715, 38), event.snapshot_hits[1])
        self.assertTupleEqual((103, 17, 728720, 18), event.snapshot_hits[2])



# coding=utf-8
# Filename: constants.py
# pylint: disable=C0103
"""
The constants used in KM3Pipe.

"""
from __future__ import division, absolute_import, print_function

#TODO: this module should be refactored soon!

import math


c = 2.99792458e8  # m/s

n_water_antares_phase = 1.3499
n_water_antares_group = 1.3797
n_water_antares = n_water_antares_group
theta_cherenkov_water_antares = math.acos(1 / n_water_antares_phase)
c_water_antares = c / n_water_antares_group

# Math
pi = math.pi
e = math.e

# Default values for time residuals
dt_window_l = -15  # ns
dt_window_h = +25  # ns



# coding=utf-8
# Filename: app.py
"""
PipeInspector

Usage:
    pipeinspector FILE
    pipeinspector (-h | --help)
    pipeinspector --version

Options:
    -h --help       Show this screen.

"""
from __future__ import division, absolute_import, print_function

import os

import urwid

from pipeinspector.gui import MainFrame
from pipeinspector.settings import UI
from km3pipe.pumps import EvtPump, DAQPump, AanetPump, CLBPump

__version__ = "1.0.0"


def handle_input(input):
    """Handle any unhandled input."""
    if input in UI.keys['escape']:
        raise urwid.ExitMainLoop

def filter_input(keys, raw):
    """Adds fancy mouse wheel functionality and VI navigation to ListBox"""
    if len(keys) == 1:
        if keys[0] in UI.keys['up']:
            keys[0] = 'up'
        elif keys[0] in UI.keys['down']:
            keys[0] = 'down'
        elif len(keys[0]) == 4 and keys[0][0] == 'mouse press':
            if keys[0][1] == 4:
                keys[0] = 'up'
            elif keys[0][1] == 5:
                keys[0] = 'down'
    return keys


def get_pump(input_file):
    extension = os.path.splitext(input_file)[1][1:]
    if extension == 'evt':
        pump = EvtPump(filename=input_file, cache_enabled=True)
    elif extension == 'dat':
        pump = DAQPump(filename=input_file)
    elif extension == 'dqd':
        pump = CLBPump(filename=input_file, cache_enabled=True)
    elif extension == 'root':
        pump = AanetPump(filename=input_file)
    else:
        raise SystemExit("No pump found for '{0}' files.".format(extension))
    return pump


def main():
    from docopt import docopt
    arguments = docopt(__doc__, version=__version__)
    input_file = arguments['FILE']
    pump = get_pump(input_file)
    main_frame = MainFrame(pump)
    #main_frame.header.set_text("Inspecting {0}".format(input_file))
    loop = urwid.MainLoop(main_frame, UI.palette,
                      input_filter=filter_input,
                      unhandled_input=handle_input)
    loop.run()


if __name__ == '__main__':
    main()



from setuptools import setup

from km3pipe import version

setup(name='km3pipe',
      version=version,
      url='http://github.com/tamasgal/km3pipe/',
      description='An analysis framework for KM3NeT',
      author='Tamas Gal',
      author_email='tgal@km3net.de',
      packages=['km3pipe', 'km3pipe.testing', 'km3pipe.pumps',
                'km3modules', 'pipeinspector'],
      include_package_data=True,
      platforms='any',
      install_requires=[
          'numpy',
          'controlhost',
          'urwid',
          'docopt',
      ],
      entry_points={
          'console_scripts': [
              'pipeinspector=pipeinspector.app:main',
          ],
      },
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
      ],
)

__author__ = 'Tamas Gal'



# coding=utf-8
# Filename: __init__.py
"""
The extemporary KM3NeT analysis framework.

"""
from __future__ import division, absolute_import, print_function

from km3pipe.__version__ import version, version_info

try:
    from km3pipe.core import Pipeline, Module, Pump, Blob, Geometry
except ImportError:
    print("Numpy is needed for KM3Pipe")

__author__ = "Tamas Gal"
__copyright__ = ("Copyright 2015, Tamas Gal and the KM3NeT collaboration "
                 "(http://km3net.org)")
__credits__ = []
__license__ = "MIT"
__version__ = version
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"



# coding=utf-8
# Filename: hardware.py
# pylint: disable=locally-disabled
"""
Classes representing KM3NeT hardware.

"""
from __future__ import division, absolute_import, print_function

__author__ = 'tamasgal'

import os

from km3pipe.tools import unpack_nfirst, split
from km3pipe.dataclasses import Point, Direction
from km3pipe.logger import logging

log = logging.getLogger(__name__)  # pylint: disable=C0103


class Detector(object):
    """The KM3NeT detector"""
    def __init__(self, filename=None):
        self.det_file = None
        self.det_id = None
        self.n_doms = None
        self.n_pmts_per_dom = None
        self.doms = {}
        self.pmts = []
        self._pmts_by_omkey = {}
        self._pmts_by_id = {}
        self._pmt_angles = []

        if filename:
            self.init_from_file(filename)

    def init_from_file(self, filename):
        """Create detector from detx file."""
        file_ext = os.path.splitext(filename)[1][1:]
        if not file_ext == 'detx':
            raise NotImplementedError('Only the detx format is supported.')
        self.open_file(filename)
        self.parse_header()
        self.parse_doms()
        self.det_file.close()

    def open_file(self, filename):
        """Create the file handler"""
        self.det_file = open(filename, 'r')

    def parse_header(self):
        """Extract information from the header of the detector file"""
        self.det_file.seek(0, 0)
        first_line = self.det_file.readline()
        self.det_id, self.n_doms = split(first_line, int)

    # pylint: disable=C0103
    def parse_doms(self):
        """Extract dom information from detector file"""
        self.det_file.seek(0, 0)
        self.det_file.readline()
        lines = self.det_file.readlines()
        try:
            while True:
                line = lines.pop(0)
                if not line:
                    continue
                try:
                    dom_id, line_id, floor_id, n_pmts = split(line, int)
                except ValueError:
                    continue
                self.n_pmts_per_dom = n_pmts
                for i in range(n_pmts):
                    raw_pmt_info = lines.pop(0)
                    pmt_info = raw_pmt_info.split()
                    pmt_id, x, y, z, rest = unpack_nfirst(pmt_info, 4)
                    dx, dy, dz, t0, rest = unpack_nfirst(rest, 4)
                    if rest:
                        log.warn("Unexpected PMT values: '{0}'".format(rest))
                    pmt_id = int(pmt_id)
                    pmt_pos = [float(n) for n in (x, y, z)]
                    pmt_dir = [float(n) for n in (dx, dy, dz)]
                    t0 = float(t0)
                    if floor_id < 0:
                        _, new_floor_id, _ = self.pmtid2omkey_old(pmt_id)
                        log.error("Floor ID is negative for PMT {0}.\n"
                                  "Guessing correct id: {1}"
                                  .format(pmt_id, new_floor_id))
                        floor_id = new_floor_id
                    #TODO: following line is here bc of the bad MC floor IDs
                    #      put it outside the for loop in future
                    self.doms[dom_id] = (line_id, floor_id, n_pmts)
                    omkey = (line_id, floor_id, i)
                    pmt = PMT(pmt_id, pmt_pos, pmt_dir, t0, i, omkey)
                    self.pmts.append(pmt)
                    self._pmts_by_omkey[(line_id, floor_id, i)] = pmt
                    self._pmts_by_id[pmt_id] = pmt
        except IndexError:
            pass

    @property
    def dom_positions(self):
        """The positions of the DOMs, taken from the PMT with the lowest ID."""
        return [pmt.pos for pmt in self._pmts_by_id.values()
                if pmt.daq_channel == 0]

    @property
    def pmt_angles(self):
        """A list of PMT directions sorted by PMT channel"""
        if not self._pmt_angles:
            pmts = self.pmts[:self.n_pmts_per_dom]
            self._pmt_angles = [pmt.dir for pmt in pmts]
        return self._pmt_angles

    @property
    def ascii(self):
        """The ascii representation of the detector"""
        header = "{det.det_id} {det.n_doms}".format(det=self)
        doms = ""
        for dom_id, (line_id, floor_id, n_pmts) in self.doms.iteritems():
            doms += "{0} {1} {2} {3}\n".format(dom_id, line_id, floor_id, n_pmts)
            for i in xrange(n_pmts):
                pmt = self._pmts_by_omkey[(line_id, floor_id, i)]
                doms += " {0} {1} {2} {3} {4} {5} {6} {7}\n".format(
                        pmt.id, pmt.pos.x, pmt.pos.y, pmt.pos.z,
                        pmt.dir.x, pmt.dir.y, pmt.dir.z,
                        pmt.t0
                        )
        return header + "\n" + doms

    def write(self, filename):
        with open(filename, 'w') as f:
            f.write(self.ascii)
        print("Detector file saved as '{0}'".format(filename))

    def pmt_with_id(self, pmt_id):
        """Get PMT with pmt_id"""
        try:
            return self._pmts_by_id[pmt_id]
        except KeyError:
            raise KeyError("No PMT found for ID: {0}".format(pmt_id))

    def pmtid2omkey(self, pmt_id):
        return self._pmts_by_id[int(pmt_id)].omkey

    def pmtid2omkey_old(self, pmt_id,
                    first_pmt_id=1, oms_per_line=18, pmts_per_om=31):
        """Convert (consecutive) raw PMT IDs to Multi-OMKeys."""
        pmts_per_line = oms_per_line * pmts_per_om
        line = ((pmt_id - first_pmt_id) // pmts_per_line) + 1
        om = oms_per_line - (pmt_id - first_pmt_id) % pmts_per_line // pmts_per_om
        pmt = (pmt_id - first_pmt_id) % pmts_per_om
        return int(line), int(om), int(pmt)



class PMT(object):
    """Represents a photomultiplier"""
    def __init__(self, id, pos, dir, t0, daq_channel, omkey):
        self.id = id
        self.pos = Point(pos)
        self.dir = Direction(dir)
        self.t0 = t0
        self.daq_channel = daq_channel
        self.omkey = omkey

    def __str__(self):
        return "PMT id:{0} pos: {1} dir: dir{2} t0: {3} DAQ channel: {4}"\
               .format(self.id, self.pos, self.dir, self.t0, self.daq_channel)

